# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains functions used in TIS.

This module defines the functions needed to perform TIS simulations
(algorithms implemented as described by van Erp et al. [TIS]_ ) with
advanced shooting moves, such as Stone Skipping and Web Throwing
(algorithms implemented as described by Riccardi et al. [SS+WT]_ ).

Methods defined here
~~~~~~~~~~~~~~~~~~~~

make_tis_step (:py:func:`.make_tis_step`)
    A method that will perform a single TIS step.

make_tis_step_ensemble (:py:func:`.make_tis_step_ensemble`)
    A method to perform a TIS step for a path ensemble. It will handle
    adding of the path to a path ensemble object.

shoot (:py:func:`.shoot`)
    A method that will perform a shooting move.

select_shoot (:py:func:`.select_shoot`)
    A method that randomly selects the shooting point.

wire_fencing (:py:func:`.wire_fencing`)
    A method that will do a wire fencing shooting move.

stone_skipping (:py:func:`.stone_skipping`)
    A method that will do a stone skipping shooting move.

web_throwing (:py:func:`.web_throwing`)
    A method that will do a web throwing shooting move.

extender (:py:func:`.extender`)
    A method that will extend a path to target interfaces.

time_reversal (:py:func:`.time_reversal`)
    A method for performing the time reversal move.

ss_wt_wf_acceptance (:py:func:`.ss_wt_wf_acceptance`)
    A method to check the acceptance of a newly generated path
    according to super detailed balance rules for Stone Skipping,
    Wire Fencing (basic and High Acceptance version) and Web Throwing.


References
~~~~~~~~~~
.. [SS+WT] E. Riccardi, O. Dahlen, T. S. van Erp,
   J. Phys. Chem. letters, 8, 18, 4456, (2017),
   https://doi.org/10.1021/acs.jpclett.7b01617

.. [TIS] T. S. van Erp, D. Moroni, P. G. Bolhuis,
   J. Chem. Phys. 118, 7762 (2003),
   https://dx.doi.org/10.1063%2F1.1562614

"""
import logging
from pyretis.core.common import (null_move,
                                 counter,
                                 relative_shoots_select,
                                 segments_counter,
                                 compute_weight,
                                 select_and_trim_a_segment,
                                 crossing_counter,
                                 priority_checker,
                                 crossing_finder,
                                 wirefence_weight_and_pick)

from pyretis.core.montecarlo import metropolis_accept_reject
from pyretis.core.path import Path, paste_paths
from pyretis.inout.restart import write_ensemble_restart

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

__all__ = ['extender',
           'make_tis',
           'make_tis_step',
           'make_tis_step_ensemble',
           'paste_paths',
           'select_shoot',
           'shoot',
           'ss_wt_wf_acceptance',
           'wire_fencing',
           'stone_skipping',
           'time_reversal',
           'web_throwing']


def make_tis(ensembles, rgen, settings, cycle):
    """Perform a TIS step in a given path ensemble.

    This function will execute the TIS steps. When used in the RETIS method,
    three different options can be given:

    1) If `relative_shoots` is given in the input settings as a list of
       probability for each ensemble to be picked, then an ensemble
       will be ranomly picked and TIS performed on. For all the
       other ensembles we again have two options based on the given
       `settings['nullmoves']`:

       a) Do a 'null move' in all other ensembles.
       b) Do nothing for all other ensembles.

       Performing the null move in an ensemble will simply just accept
       the previously accepted path in that ensemble again.

    2) If `relative_shoots` is not given in the input settings, TIS moves
       will be executed for all path ensembles.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the TIS method
        with their properties.
        They contain:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    rgen : object like :py:class:`.RandomGenerator`
        This is the random generator that will be used.
    settings : dict
        This dictionary contains the TIS settings.
    cycle : int
        The current cycle number.

    Yields
    ------
    out : dict
        This dictionary contains the results of the TIS move(s).

    """
    relative = settings.get('retis', {}).get('relative_shoots', None)
    nullmoves = settings.get('retis', {}).get('nullmoves', False)

    if relative is not None:
        idx, ensemble = relative_shoots_select(ensembles, rgen, relative)
        accept, trial, status = make_tis_step_ensemble(
            ensemble, settings['ensemble'][idx], cycle
        )
        result = {
            'ensemble_number': idx,
            'mc-move': trial.get_move(),
            'status': status,
            'trial': trial,
            'accept': accept
        }
        yield result
        # Do null moves in the other ensembles, if requested:
        if nullmoves:
            for ensemble in ensembles:
                other = ensemble['path_ensemble'].ensemble_number
                if other != idx:
                    accept, trial, status = null_move(ensemble, cycle)
                    result = {
                        'ensemble_number': other,
                        'mc-move': 'nullmove',
                        'status': status,
                        'trial': trial,
                        'accept': accept,
                    }
                    if 0 == cycle % settings['ensemble'][idx].get(
                            'output', {}).get('restart-file', 1):
                        write_ensemble_restart(ensemble, settings)
                    yield result
    else:  # Do TIS in all ensembles:
        prio_skip = priority_checker(ensembles, settings)
        for i, (ens, e_set) in enumerate(zip(ensembles, settings['ensemble'])):
            if prio_skip[i]:
                # Here we skip the ensembles with higher cycle numbers.
                yield
            else:
                accept, trial, status = make_tis_step_ensemble(ens,
                                                               e_set,
                                                               cycle)
                result = {
                    'ensemble_number': ens['path_ensemble'].ensemble_number,
                    'mc-move': trial.get_move(),
                    'status': status,
                    'trial': trial,
                    'accept': accept
                }
                yield result


def make_tis_step_ensemble(ensemble, settings, cycle):
    """Perform a TIS step for a given path ensemble.

    This function will run `make_tis_step` for the given path_ensemble.
    It will handle adding of the path. This function is intended for
    convenience when working with path ensembles. If we are using the
    path ensemble ``[0^-]`` then the start condition should be 'R' for
    right.

    Parameters
    ----------
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.

    settings : dict
        This dictionary contains the ensemble settings.
    cycle : int
        The current cycle number.

    Returns
    -------
    out[0] : boolean
        True if the new path can be accepted, False otherwise.
    out[1] : object like :py:class:`.PathBase`
        The generated path.
    out[2] : string
        The status of the path.

    """
    path_ensemble = ensemble['path_ensemble']
    engine = ensemble['engine']
    start_cond = path_ensemble.start_condition
    logger.info('TIS move in: %s', path_ensemble.ensemble_name)
    engine.exe_dir = path_ensemble.directory['generate']
    accept, trial, status = make_tis_step(ensemble,
                                          settings['tis'],
                                          start_cond)

    # If we are exploring, we accept everything.
    if settings['simulation']['task'] == 'explore':
        if trial.length > 2:
            accept, status = True, 'EXP'
            trial.status = 'EXP'

    if accept:
        logger.info('The move was accepted!')
    else:
        logger.info('The move was rejected! (%s)', status)

    path_ensemble.add_path_data(trial, status, cycle=cycle)
    if cycle % settings.get('output', {}).get('restart-file', 1) == 0:
        write_ensemble_restart(ensemble, settings)
    return accept, trial, status


def make_tis_step(ensemble, tis_settings, start_cond):
    """Perform a TIS step and generate a new path.

    The new path will be generated from the input path, either by
    performing a time-reversal move or by a shooting move. This is
    determined pseudo-randomly by drawing a random number from a
    uniform distribution using the given random generator.

    Parameters
    ----------
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `path`: object like :py:class:`.PathBase`
          This is the input path which will be used for generating a
          new path.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_settings : dict
        This dictionary contains the settings for the TIS method. Here we
        explicitly use:

        * `freq`: float, the frequency of how often we should do time
          reversal moves.
        * `shooting_move`: string, the label of the shooting move to perform.

    start_cond : string
        The starting condition for the path. This is determined by the
        ensemble we are generating for - it is 'R'ight or 'L'eft.

    Returns
    -------
    out[0] : boolean
        True if the new path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        The generated path.
    out[2] : string
        The status of the path.

    """
    interfaces = ensemble['interfaces']
    order_function = ensemble['order_function']
    path_ensemble = ensemble['path_ensemble']
    rgen = ensemble['rgen']
    last_path = path_ensemble.last_path
    # Get weights
    moves = ['tr', 'mirror', 'targetswap', 'sh']
    weights = [tis_settings.get('freq', 0),
               tis_settings.get('mirror_freq', 0),
               tis_settings.get('target_freq', 0),
               0]

    # Mirror and target swap only makes sense for ensemble 000
    if path_ensemble.ensemble_number != 0:
        weights[moves.index('mirror')] = 0
        weights[moves.index('targetswap')] = 0

    # Add sh weight (make sure it is non-negative)
    weights[-1] = max(1-sum(weights), 0)

    if sum(weights) > 1:
        msg = "move probabilities sum to more than 1. Got the following "
        msg += f"probabilities: {list(zip(moves, weights))}"
        raise ValueError(msg)

    choice = rgen.choice(moves, 1, p=weights)[0]
    if choice == 'tr':
        logger.info('Performing a time reversal move')
        accept, new_path, status = time_reversal(last_path, order_function,
                                                 interfaces, start_cond)
    elif choice == 'sh':
        logger.info('Selecting the shooting move.')
        accept, new_path, status = select_shoot(ensemble, tis_settings,
                                                start_cond)
    elif choice == 'mirror':
        logger.info("Selecting the mirror move")
        accept, new_path, status = mirror(ensemble)

    elif choice == 'targetswap':
        logger.info("Selecting the target swap move")
        accept, new_path, status = target_swap(ensemble, tis_settings)

    else:
        raise RuntimeError(f"Illegal move choice: {choice}")

    return accept, new_path, status


def select_shoot(ensemble, tis_settings, start_cond):
    """Select the shooting move to generate a new path.

    The new path will be generated from the input path, either by
    performing a normal shooting or web-throwing. This is
    determined pseudo-randomly by drawing a random number from a
    uniform distribution using the given random generator.

    Parameters
    ----------
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `path`: object like :py:class:`.PathBase`
          This is the input path which will be used for generating a
          new path.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_settings : dict
        This dictionary contains the settings for the TIS method. Here we
        explicitly use:

        * `freq`: float, the frequency of how often we should do time
          reversal moves.
        * `shooting_move`: string, the label of the shooting move to perform.

    start_cond : string
        The starting condition for the path. This is determined by the
        ensemble we are generating for - it is 'R'ight or 'L'eft.

    Returns
    -------
    out[0] : boolean
        True if the new path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        The generated path.
    out[2] : string
        The status of the path.

    """
    shooting_move = tis_settings.get('shooting_move', 'sh')

    if shooting_move == 'wt':
        logger.info('Performing a Web Throwing move')
        accept, new_path, status = web_throwing(ensemble, tis_settings)

    elif shooting_move == 'ss':
        logger.info('Performing a Stone Skipping move')
        accept, new_path, status = stone_skipping(ensemble, tis_settings,
                                                  start_cond)
    elif shooting_move == 'wf':
        logger.info('Performing a Wire Fencing move')
        accept, new_path, status = wire_fencing(ensemble, tis_settings,
                                                start_cond)
    elif shooting_move == '00':
        logger.info('Performing a null move')
        accept, new_path, status = null_move(ensemble, 0)
    else:
        logger.info('Performing a shooting move')
        accept, new_path, status = shoot(ensemble, tis_settings, start_cond)

    return accept, new_path, status


def time_reversal(path, order_function, interfaces, start_condition):
    """Perform a time-reversal move.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be used for generating a
        new path.
    order_function : object like :py:class:`.OrderParameter`
        The class used for obtaining the order parameter(s).
    interfaces : list/tuple of floats
        These are the interface positions of the form
        ``[left, middle, right]``.
    start_condition : string
        The starting condition, 'L'eft or 'R'ight.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        The time reversed path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        `.path._STATUS`.

    """
    new_path = path.reverse(order_function)
    start, _, _, _ = new_path.check_interfaces(interfaces)
    # Explicitly set how this was generated. Keep the old if loaded.
    if path.get_move() == 'ld':
        new_path.set_move('ld')
    else:
        new_path.generated = ('tr', 0, 0, 0)

    if start in set(start_condition):
        accept = True
        status = 'ACC'
    else:
        accept = False
        status = 'BWI'  # Backward trajectory end at wrong interface.
    new_path.status = status
    return accept, new_path, status


def mirror(ensemble):
    """Perform a mirror move.

    This function performs a mirror move by mirroring the order function in
    place. It then calculates the new order parameter for each phase point in
    the path. The generated path is then returned. If the path is loaded, then
    the newly generated path will also carry the generate-attribute 'ld'.
    This move is always acceptable, so the status is set to 'ACC', and the
    bool accept is set to True.

    Parameters
    ----------
    ensemble : dictionary of objects

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        The generated path.
    out[2] : string
        Status of the path,

    """
    order_function = ensemble['order_function']
    # This also mirrors the ensemble['of'] inplace
    order_function.mirror()
    path = ensemble['path_ensemble'].last_path
    new_path = path.copy()
    # Calculate the new orderparameter
    for phasepoint in new_path.phasepoints:
        phasepoint.order = order_function.calculate(phasepoint)

    # This does not fix loaded paths
    if path.get_move() == 'ld':
        new_path.set_move('ld')
    else:
        new_path.generated = ('mr', 0, 0, 0)

    # This is an always accept move
    accept = True
    status = 'ACC'
    new_path.status = status
    return accept, new_path, status


def get_target_swap_choices(allowed_indices, path, ensemble):
    """Get the valid choices for the target swap move.

    Parameters
    ----------
    allowed_indices : list of integers
        The allowed indices for the target swap.
    path : object like :py:class:`.PathBase`
        The path to use for the target swap.
    ensemble : dictionary of objects
        The ensemble to use for the target swap.

    Returns
    -------
    choices : list of tuples
        The choices for the target swap.
        tuple = (index, frame)

    """
    order_function = ensemble['order_function']
    left, _, right = tuple(ensemble['interfaces'])
    choices = []
    old_idx = order_function.index
    for idx in allowed_indices:
        # swap out index on the order function
        order_function.index = idx
        for frame, phasepoint in enumerate(path.phasepoints):
            if left <= order_function.calculate(phasepoint)[0] < right:
                choices.append((idx, frame))
    order_function.index = old_idx
    return choices


def target_swap_backward(choice, choices, old_path, new_path, ensemble):
    """Perform the backward part of the target swap move.

    This function loops backward from a selected frame (the "trial point")
    until it detects a frame where the new target has a position outside a
    certain interval. The number of steps taken in this process is recorded as
    `n_n`. If there are not enough frames, it propagates backwards to
    generate more.

    Parameters
    ----------
    choice : tuple
        The choice to use for the target swap.
        tuple = (index, frame)
    choices : list of tuples
        The choices for the target swap.
        tuple = (index, frame)
    old_path : object like :py:class:`.PathBase`
        The path to use for the target swap.
    new_path : object like :py:class:`.PathBase`
        The new path to use for the target swap.
    ensemble : dictionary of objects
        The ensemble to use for the target swap.

    Returns
    -------
    n_n : integer
        The number of valid frames in the new path where a swap could
        potentially occur.
    n_o : integer
        The number of valid frames in the old path where a swap could
        potentially occur.

    """
    order_function = ensemble['order_function']
    engine = ensemble['engine']
    old_idx = order_function.index
    order_function.index = choice[0]
    n_n = 0
    first_frame = choice[1]
    n_o = 0

    # First loop back:
    for i in range(choice[1], -1, -1):
        if (choice[0], i) not in choices:
            break
        n_n += 1
        first_frame = i

        # Prevent double frames if we need to extend
        if i != 0:
            phasepoint = old_path.phasepoints[i].copy()
            phasepoint.order = order_function.calculate(phasepoint)
            new_path.append(phasepoint)

    if first_frame != 0:
        # We know we can append at least 1 more frame that would end up in a
        # state
        extra_frame = first_frame - 1
        logger.debug("Found path starting at frame %s, skipping backward MD",
                     extra_frame)
        phasepoint = old_path.phasepoints[extra_frame].copy()
        phasepoint.order = order_function.calculate(phasepoint)
        new_path.append(phasepoint)
        # Not all old_points are valid
        if extra_frame != 0:
            n_o -= extra_frame-1
            logger.debug("n_old updated from %s to %s",
                         n_o+(extra_frame-1), n_o)

    else:
        logger.debug("Need more frames, propagating backwards")
        ensemble['system'] = old_path.phasepoints[0].copy()
        engine.propagate(new_path, ensemble, reverse=True)
    order_function.idx = old_idx
    return n_n, n_o


def target_swap_forward(choice, choices, old_path, new_path, ensemble):
    """Perform the forward part of the target swap move.

    This function loops forward from a selected frame (the "trial point") until
    it detects a frame where the new target has a position outside a certain
    interval. The number of steps taken in this process is recorded as `n_n`.
    If there are not enough frames, it propagates forwards to generate more.

    Parameters
    ----------
    choice : tuple
        The choice to use for the target swap.
        tuple = (index, frame)
    choices : list of tuples
        The choices for the target swap.
        tuple = (index, frame)
    old_path : object like :py:class:`.PathBase`
        The path to use for the target swap.
    new_path : object like :py:class:`.PathBase`
        The new path to use for the target swap.
    ensemble : dictionary of objects
        The ensemble to use for the target swap.

    Returns
    -------
    n_n : integer
        The number of valid frames in the new path where a swap could
        potentially occur.
    n_o : integer
        The number of valid frames in the old path where a swap could
        potentially occur.

    """
    order_function = ensemble['order_function']
    engine = ensemble['engine']
    old_idx = order_function.index
    order_function.index = choice[0]
    n_n = 0
    last_frame = choice[1]
    n_o = 0

    for i in range(choice[1]+1, old_path.length, 1):
        if (choice[0], i) not in choices:
            break
        n_n += 1
        last_frame = i
        # Prevent last point if we need to extend
        if i != old_path.length-1:
            phasepoint = old_path.phasepoints[i].copy()
            phasepoint.order = order_function.calculate(phasepoint)
            new_path.append(phasepoint)

    if last_frame < old_path.length-1:
        # We know we can append at least one frame that will result in a valid
        # path
        extra_frame = last_frame + 1
        logger.debug("Found path ending at frame %s, skipping forward MD",
                     extra_frame)
        phasepoint = old_path.phasepoints[extra_frame].copy()
        phasepoint.order = order_function.calculate(phasepoint)
        new_path.append(phasepoint)
        # Not all old_points are valid
        # -2 because we already dismissed the last frame
        # Need to catch when we append the last frame here as we don't want to
        # add that one
        if extra_frame != old_path.length-1:
            n_o -= (old_path.length-2)-extra_frame
            logger.debug("n_old updated from %s to %s",
                         n_o+((old_path.length-2)-extra_frame), n_o)
    else:
        logger.debug("Need more frames, propagating forwards")
        ensemble['system'] = old_path.phasepoints[-1].copy()
        engine.propagate(new_path, ensemble, reverse=False)

    order_function.index = old_idx
    return n_n, n_o


def target_swap(ensemble, tis_settings):
    """Perform a target swap move.

    This function attempts a target swap move starting from a given index to
    allowed indices. It finds a trial point from which a new path is generated.
    It records Z(o->n) and Z(n->o), which are counts of valid frames where a
    target swap could potentially occur (from old to new, and from new to old,
    respectively). These are used to calculate the acceptance probability.

    Once a trial point is chosen, it loops forwards and backwards in the
    trajectory of the old path (along the new choice index) to construct a new
    path. This may include MD integration if necessary. If there is no valid
    swap, it returns with status 'TSS'.

    Parameters
    ----------
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
            This is the path ensemble to perform the target swap move for.
        * `order_function`: object like :py:class:`.OrderParameter`
            The class used for obtaining the order parameter(s).

    tis_settings : dict
        This dictionary contains the settings for the TIS method. Here we
        explicitly use:

        * `freq`: float, the frequency of how often we should do time
          reversal moves.
        * `shooting_move`: string, the label of the shooting move to perform.

    Returns
    -------
    tuple : A tuple containing:
        - accept (bool): Whether the move was accepted or not.
        - new_path (object like :py:class:`.PathBase`): The new path.
        - status (str): The status of the move.
          Can be 'ACC', 'TSS', 'BTX', or 'FTX'

    Raises
    ------
    ValueError : If the current order parameter index is not part of the target
        indices.

    """
    # Grab required variables
    target_indices = tis_settings['target_indices']
    order_f = ensemble['order_function']
    old_index = order_f.index
    if old_index not in target_indices:
        raise ValueError(f"Current orderparameter index: {old_index}, is not "
                         f"part of the target indices: {target_indices}.")
    old_path = ensemble['path_ensemble'].last_path
    maxlen = tis_settings['maxlength']
    new_path = old_path.empty_path(maxlen=maxlen)
    rgen = ensemble['rgen']

    allowed_indices = [i for i in target_indices if i != old_index]

    logger.debug("Attempt a target swap starting from index %s "
                 "to the allowed indices: %s",
                 old_index, allowed_indices)

    # 2 Find trial point
    choices = get_target_swap_choices(allowed_indices=allowed_indices,
                                      path=old_path,
                                      ensemble=ensemble)
    # If there is no valid swap
    if len(choices) == 0:
        logger.info("Found no valid swapping frames")
        accept = False
        status = 'TSS'
        new_path = old_path.copy()
        new_path.generated = ('ts', 0, 0, 0)
        new_path.status = status
        return accept, new_path, status
    choice_idx = rgen.choice(len(choices), 1)[0]

    choice = choices[choice_idx]
    logger.debug("Choose %s out of %s.", choice, choices)
    logger.info("Attempting target swap from index %s to index %s",
                old_index, choice[0])

    # Record Z(o->n)
    z_o_n = len(choices)
    # Figure out n_o
    n_n = 0
    # default n_o is the old path min the 2 frames in the states
    n_o = old_path.length-2

    # First loop backwards
    diff_n_n, diff_n_o = target_swap_backward(choice=choice,
                                              choices=choices,
                                              old_path=old_path,
                                              new_path=new_path,
                                              ensemble=ensemble)
    n_n += diff_n_n
    n_o += diff_n_o

    new_path = new_path.reverse()
    # Needed because path.reverse does not clone attributes
    new_path.maxlen = maxlen
    if new_path.length >= maxlen:
        order_f.index = old_index
        logger.debug("Backward trajectory to long")
        accept = False
        status = 'BTX'
        new_path.generated = ('ts', 0, 0, 0)
        new_path.status = status
        return accept, new_path, status

    # Loop forward:
    diff_n_n, diff_n_o = target_swap_forward(choice=choice,
                                             choices=choices,
                                             old_path=old_path,
                                             new_path=new_path,
                                             ensemble=ensemble)
    n_n += diff_n_n
    n_o += diff_n_o
    if new_path.length >= maxlen:
        order_f.index = old_index
        logger.debug("Forward trajectory to long")
        accept = False
        status = 'FTX'
        new_path.generated = ('ts', 0, 0, 0)
        new_path.status = status
        return accept, new_path, status

    # Need to calculate Z(n->o)
    allowed_indices = [i for i in target_indices if i != choice[0]]
    new_choices = get_target_swap_choices(allowed_indices=allowed_indices,
                                          path=new_path,
                                          ensemble=ensemble)

    z_n_o = len(new_choices)

    logger.debug("Going to calculate acceptance\n"
                 "Old choices: %s\n"
                 "n_n: %s, Z_o_n: %s\n"
                 "New choices: %s\n"
                 "n_o: %s, Z_n_o: %s\n",
                 choices,
                 n_n, z_o_n,
                 new_choices,
                 n_o, z_n_o)

    acc = min(1, n_o/n_n*z_o_n/z_n_o)
    ran_num = rgen.rand()[0]
    logger.info("prob = %s, random = %s", acc, ran_num)
    if ran_num >= acc:
        order_f.index = old_index
        logger.debug("Failed acceptance")
        accept = False
        status = "TSA"
        new_path.generated = ('ts', 0, 0, 0)
        new_path.status = status
        return accept, new_path, status

    order_f.index = choice[0]
    logger.debug("Accepted target swap")
    accept = True
    status = "ACC"
    new_path.generated = ('ts', 0, 0, 0)
    new_path.status = status
    return accept, new_path, status


def prepare_shooting_point(path, ensemble, tis_settings):
    """Select and modify velocities for a shooting move.

    This method will randomly select a shooting point from a given
    path and modify its velocities.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be used for generating a
        new path.
    ensemble : dictionary of objects
        It contains:

        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list/tuple of floats
          These are the interface positions of the form
          ``[left, middle, right]``.

    tis_settings : dict
        This contains the settings for TIS. Here, we use the
        settings which dictates how we modify the velocities.
        * `aimless`: boolean, is the shooting aimless or not?

    Returns
    -------
    out[0] : object like :py:class:`.System`
        The shooting point with modified velocities.
    out[1] : integer
        The index of the shooting point in the original path.
    out[2] : float
        The change in kinetic energy when modifying the velocities.

    """
    shooting_point, idx = path.get_shooting_point(
        criteria=tis_settings.get('shooting_move', 'rnd'),
        interfaces=ensemble.get('interfaces'))
    engine = ensemble['engine']
    orderp = shooting_point.order
    logger.info('Shooting from order parameter/index: %f, %d', orderp[0], idx)
    # Copy the shooting point, so that we can modify velocities without
    # altering the original path:
    shooting_copy = shooting_point.copy()
    ensemble['system'] = shooting_copy
    # Modify the velocities:
    dek, _, = engine.modify_velocities(
        ensemble,
        {'sigma_v': tis_settings['sigma_v'],
         'aimless': tis_settings['aimless'],
         'zero_momentum': tis_settings['zero_momentum'],
         'rescale': tis_settings['rescale_energy']})
    orderp = engine.calculate_order(ensemble)
    shooting_copy.order = orderp
    return shooting_copy, idx, dek


def one_step_crossing(ensemble, interface):
    """Create a path of one step and check the crossing with the interface.

    This function will do a single step to try to cross the interface.
    Note that a step might involve several substeps, depending on the
    input file selection/sampling strategy.
    This task is the Achilles` Heel of Stone Skipping. If not wisely
    done, a large number of attempts to cross the interface
    in one step will be done, destroying the sampling efficiency.


    Parameters
    ----------
    ensemble : dictionary of objects
        It contains:

        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
    interface : float
        This is the interface position to be crossed.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.

    """
    # The trial path we need to generate. Note, 1 step = 2 points
    trial_path = Path(rgen=ensemble['rgen'], maxlen=2)
    interfaces = [-float("inf"), float("inf"), float("inf")]
    sub_ensemble = {'interfaces': interfaces,
                    'system': ensemble['system'],
                    'order_function': ensemble['order_function']}
    ensemble['engine'].propagate(trial_path, sub_ensemble)
    if crossing_counter(trial_path, interface) == 0:
        return False, trial_path

    return True, trial_path


def check_kick(shooting_point, interfaces, trial_path, rgen, dek,
               tis_settings):
    """Check the modification of the shooting point.

    After generating velocities for a shooting point, we
    do some additional checking to see if the shooting point is
    acceptable.

    Parameters
    ----------
    shooting_point : object like :py:class:`.System`
        The shooting point with modified velocities.
    interfaces : list of floats
        The interfaces used for TIS, in the format
        ``[left, middle, right]``.
    trial_path : object like :py:class:`.PathBase`
        The path we are currently generating.
    rgen : object like :py:class:`.RandomGenerator`
        This is the random generator that will be used to check if
        we accept the shooting point based on the change in kinetic
        energy.
    dek : float
        The change in kinetic energy when modifying the velocities.
    tis_settings : dict
        This contains the settings for TIS.

    Returns
    -------
    out : boolean
        True if the kick was OK, False otherwise.

    """
    # 1) Check if the kick was too violent:
    left, _, right = interfaces
    if 'exp' in tis_settings.get('shooting_move', {}):
        return True
    if not left <= shooting_point.order[0] < right:
        # Shooting point was velocity dependent and was kicked outside
        # of boundaries when modifying velocities.
        trial_path.append(shooting_point)
        trial_path.status = 'KOB'
        return False
    # 2) If the kick is not aimless, we check if we reject it or not:
    if not tis_settings['aimless']:
        accept_kick = metropolis_accept_reject(rgen, shooting_point, dek)
        # If one wish to implement a bias call, this can be done here.
        if not accept_kick:
            trial_path.append(shooting_point)
            trial_path.status = 'MCR'  # Momenta Change Rejection.
            return False
    return True


def shoot_backwards(path_back, trial_path, ensemble,
                    tis_settings, start_cond):
    """Shoot in the backward time direction.

    Parameters
    ----------
    path_back : object like :py:class:`.PathBase`
        The path we will fill with phase points from the propagation.
    trial_path : object like :py:class:`.PathBase`
        The current trial path generated by the shooting.
    ensemble : dictionary of objects
        It contains:

        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `interfaces`: list/tuple of floats
          These are the interface positions of the form
          ``[left, middle, right]``.
        * `system`: object like :py:class:`.System`
           The system that originates the path.

    tis_settings : dict
        This contains the settings for TIS.
    start_cond : string
        The starting condition for the current ensemble, 'L'eft or
        'R'ight.

    Returns
    -------
    out : boolean
        True if the backward path was generated successfully, False
        otherwise.

    """
    logger.debug('Propagating backwards for the shooting move.')
    path_back.time_origin = trial_path.time_origin
    engine = ensemble['engine']
    success_back, _ = engine.propagate(path_back, ensemble, reverse=True)
    if not success_back:
        # Something went wrong, most probably the path length was exceeded.
        trial_path.status = 'BTL'  # BTL = backward trajectory too long.
        # Add the failed path to trial path for analysis:
        trial_path += path_back
        if path_back.length >= tis_settings['maxlength'] - 1:
            # BTX is backward trajectory longer than maximum memory.
            trial_path.status = 'BTX'
        return False
    # Backward seems OK so far, check if the ending point is correct:
    left, _, right = ensemble['interfaces']
    if path_back.get_end_point(left, right) not in set(start_cond):
        # Nope, backward trajectory end at wrong interface.
        trial_path += path_back  # Store path for analysis.
        trial_path.status = 'BWI'
        return False
    return True


def shoot(ensemble, tis_settings, start_cond, shooting_point=None):
    """Perform a shooting-move.

    This function will perform the shooting move from a randomly
    selected time-slice.

    Parameters
    ----------
    ensemble: dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_settings : dict
        This contains the settings for TIS:

        * `aimless`: boolean, is the shooting aimless or not?
        * `allowmaxlength`: boolean, should paths be allowed to reach
          maximum length?
        * `maxlength`: integer, maximum allowed length of paths.

    start_cond : string or tuple of strings
        The starting condition for the current ensemble, 'L'eft or
        'R'ight. ('L', 'R'), the default option, implies no directional
        difference.
    shooting_point: object like :py:class:`.System`, optional
        If given, it is the shooting point from which the path is generated.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        :py:const:`.path._STATUS`.

    """
    path_ensemble = ensemble['path_ensemble']
    path = path_ensemble.last_path
    interfaces = ensemble['interfaces']
    trial_path = path.empty_path()  # The trial path we will generate.
    if shooting_point is None:
        shooting_point, idx, dek = prepare_shooting_point(
            path, ensemble, tis_settings
        )
        kick = check_kick(shooting_point, interfaces, trial_path, path.rgen,
                          dek, tis_settings)
    else:
        kick = True
        idx = getattr(shooting_point, 'idx', 0)

    # Store info about this point, just in case we have to return
    # before completing a full new path:
    trial_path.generated = ('sh', shooting_point.order[0], idx, 0)
    trial_path.time_origin = path.time_origin + idx
    # We now check if the kick was OK or not:
    if not kick:
        return False, trial_path, trial_path.status
    # OK: kick was either aimless or it was accepted by Metropolis
    # we should now generate trajectories, but first check how long
    # it should be (if the path comes from a load, it is assumed to not
    # respect the detail balance anyway):
    if path.get_move() == 'ld' or tis_settings['allowmaxlength']:
        maxlen = tis_settings['maxlength']
    else:
        maxlen = min(int((path.length - 2) / path.rgen.rand()[0]) + 2,
                     tis_settings['maxlength'])
    # Since the forward path must be at least one step, the maximum
    # length for the backward path is maxlen-1.
    # Generate the backward path:
    path_back = path.empty_path(maxlen=maxlen - 1)
    # Set ensemble state to the selected shooting point:
    ensemble['system'] = shooting_point.copy()
    if not shoot_backwards(path_back, trial_path, ensemble,
                           tis_settings, start_cond):
        return False, trial_path, trial_path.status

    # Everything seems fine, now propagate forward.
    # Note that the length of the forward path is adjusted to
    # account for the fact that it shares a point with the backward
    # path (i.e. the shooting point). The duplicate point is just
    # counted once when the paths are merged by the method
    # `paste_paths` by setting `overlap=True` (which indicates that
    # the forward and backward paths share a point).
    path_forw = path.empty_path(maxlen=maxlen - path_back.length + 1)
    logger.debug('Propagating forwards for shooting move...')
    # Set ensemble state to the selected shooting point:
    # change the system state.
    ensemble['system'] = shooting_point.copy()
    success_forw, _ = ensemble['engine'].propagate(path_forw, ensemble,
                                                   reverse=False)
    path_forw.time_origin = trial_path.time_origin
    # Now, the forward propagation could have failed by exceeding the
    # maximum length for the forward path. However, it could also fail
    # when we paste together so that the length is larger than the
    # allowed maximum. We paste first and ask later:
    trial_path = paste_paths(path_back, path_forw, overlap=True,
                             maxlen=tis_settings['maxlength'])

    # Also update information about the shooting:
    trial_path.generated = ('sh', shooting_point.order[0], idx,
                            path_back.length - 1)
    trial_path.weight = 1.
    if not success_forw:
        trial_path.status = 'FTL'
        # If we reached this point, the backward path was successful,
        # but the forward was not. For the case where the forward was
        # also successful, the length of the trial path cannot exceed
        # the maximum length given in the TIS settings. Thus we only
        # need to check this here, i.e. when given that the backward
        # was successful and the forward not:
        if trial_path.length == tis_settings['maxlength']:
            trial_path.status = 'FTX'  # exceeds "memory".

    # Deal with the rejections for path properties.
    # Make sure we did not hit the left interface on {0-} if this is disallowed
    # Cases:
    # Permeability=true: starting/ending at L is allowed in [0-]
    # RETIS: [0-] is the only ensemble that allows paths starting R
    # (RE)PPTIS: starting/ending at L and R is allowed for body ensembles.
    #            exception: [0^{+-}'] is not allowed to start AND end at R.
    #                       As this ensemble has must_cross_M=True, and L=M,
    #                       This is automatically taken care of.
    elif ('L' not in set(path_ensemble.start_condition) and
            'L' in trial_path.check_interfaces(interfaces)[:2]):
        trial_path.status = '0-L'

    # Last check - Did we cross the middle interface?
    # if we are using PPTIS or REPPTIS
    elif getattr(path_ensemble, 'must_cross_M', False) and \
            (not trial_path.check_interfaces(interfaces)[-1][1] or
             not (trial_path.check_interfaces(interfaces)[-1][0] or
                  trial_path.check_interfaces(interfaces)[-1][2])):
        # not for [0-]
        # detect if: minorder < middle interf <= maxorder
        # or
        # if we do cross middle interface, the
        # path still has to come from left or right
        # so we need cross[0] or cross[2] to be true
        trial_path.status = 'NCR'

    # if we are not using PPTIS or REPPTIS
    # Special treatment if paths can start everywhere
    elif set(('R', 'L')) != set(path_ensemble.start_condition) and \
            not trial_path.check_interfaces(interfaces)[-1][1]:
        # No, we did not cross the middle interface:
        trial_path.status = 'NCR'
    else:  # If nothing went wrong, then...
        trial_path.status = 'ACC'
        return True, trial_path, trial_path.status

    return False, trial_path, trial_path.status


def wire_fencing(ensemble, tis_settings, start_cond):
    """Perform a wire_fencing move.

    This function will perform the non-famous wire fencing move
    from an initial path.

    Parameters
    ----------
    ensemble: dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_settings : dict
        This contains the settings for TIS. Keys used here:

        * `aimless`: boolean, is the shooting aimless or not?
        * `allowmaxlength`: boolean, should paths be allowed to reach
          maximum length?
        * `maxlength`: integer, maximum allowed length of paths.
        * `high_accept`: boolean, the option for High Acceptance WF.

    start_cond : string
        The starting condition for the current ensemble, 'L'eft or
        'R'ight.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        :py:const:`.path._STATUS`.

    """
    trial_path = ensemble['path_ensemble'].last_path
    engine = ensemble['engine']
    wf_int = [ensemble['interfaces'][1], ensemble['interfaces'][1],
              tis_settings.get('interface_cap', ensemble['interfaces'][2])]
    n_frames, new_segment = wirefence_weight_and_pick(trial_path, wf_int[0],
                                                      wf_int[2],
                                                      return_seg=True)

    # This is probably a too strong condition. It helps for [0^-] but it might
    # hinder implementation problems or bad sampling.
    if n_frames == 0:
        logger.warning('Wire fencing move not usable. N frames of Path = 0')
        logger.warning('between interfaces %s and %s.', wf_int[0], wf_int[-1])
        return False, trial_path, 'NSG'
    sub_ens = {'interfaces': wf_int, 'engine': engine,
               'rgen': ensemble['rgen'],
               'order_function': ensemble['order_function'],
               'path_ensemble': ensemble['path_ensemble']}
    sub_settings = tis_settings.copy()
    sub_settings['allowmaxlength'] = True
    succ_seg = 0

    for i in range(tis_settings['n_jumps']):
        logger.debug('Trying a new web with Wire Fencing, jump %i', i)
        # Select the shooting point:
        sh_pt, _, _ = prepare_shooting_point(new_segment, sub_ens,
                                             sub_settings)
        engine.dump_phasepoint(sh_pt, str(counter()) + '_wf_shoot')

        success, trial_seg, status = shoot(sub_ens,
                                           sub_settings,
                                           ('L', 'R'),
                                           sh_pt)
        start, end, _, _ = trial_seg.check_interfaces(wf_int)
        logger.info('Jump %s, len %s, status %s, intf: %s %s',
                    i, trial_seg.length, status, start, end)
        if not success:
            # This handles R to R (start_cond = L) paths. Counter + 1, no ups.
            logger.debug('Wire Fencing Fail.')
        else:
            logger.debug('Acceptable Wire Fence link.')
            succ_seg += 1
            new_segment = trial_seg.copy()

    if succ_seg == 0:
        # No usable segments were generated.
        trial_path.status = 'NSG'
        success = False
    else:
        success, trial_path, _ = extender(new_segment, ensemble,
                                          tis_settings, start_cond)
    if success:
        success, trial_path = ss_wt_wf_acceptance(trial_path, ensemble,
                                                  tis_settings,
                                                  start_cond)

    trial_path.generated = ('wf', sh_pt.order[0], succ_seg, trial_path.length)

    logger.debug('WF move %s', trial_path.status)
    if not success:
        return False, trial_path, trial_path.status

    # This might get triggered when accepting 0-L paths.
    left, _, right = ensemble['interfaces']
    assert start_cond == trial_path.get_start_point(left, right), \
        'WF: Path has an implausible start.'

    trial_path.status = 'ACC'
    return True, trial_path, trial_path.status


def ss_wt_wf_acceptance(trial_path, ensemble, tis_settings,
                        start_cond='L'):
    """Weights, possibly reverses and accept/rejects generated SS/WT/WF paths.

    Parameters
    ----------
    trial_path : object like :py:class:`.PathBase`
        This is the new path that will obtain weights, and might be reversed
        and accepted.
    ensemble : dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces` : list/tuple of floats
          These are the interface positions of the form
          ``[left, middle, right]``.

    tis_settings : dict
        This contains the settings for TIS. KEys used here;

        * `high_accept` : boolean, the option for High Acceptance SS/WF.
        * `shooting_move` : string, the label of the shooting move to perform.

    start_cond : string, optional
        The starting condition for the current ensemble, 'L'eft or 'R'ight.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the weighed and possibly reversed path.

    """
    intf = list(ensemble['interfaces'])
    move = tis_settings['shooting_move']

    if move == 'wt' or not tis_settings.get('high_accept', False):
        trial_path.weight = 1.
    else:
        if move == 'wf':
            intf[2] = tis_settings.get('interface_cap', intf[2])
        trial_path.weight = compute_weight(trial_path, intf, move)
        if start_cond != trial_path.get_start_point(intf[0], intf[2]):
            trial_path = trial_path.reverse(ensemble['order_function'])

    success = ss_wt_wf_metropolis_acc(trial_path, ensemble, tis_settings,
                                      start_cond)

    return success, trial_path


def stone_skipping(ensemble, tis_settings, start_cond):
    """Perform a stone_skipping move.

    This function will perform the famous stone skipping move
    from an initial path.

    Parameters
    ----------
    ensemble: dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `path`: object like :py:class:`.PathBase`
          This is the input path which will be used for generating a
          new path.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_settings : dict
        This contains the settings for TIS. Keys used here:

        * `aimless`: boolean, is the shooting aimless or not?
        * `allowmaxlength`: boolean, should paths be allowed to reach
          maximum length?
        * `maxlength`: integer, maximum allowed length of paths.
        * `high_accept`: boolean, the option for High Acceptance SS.

    start_cond : string
        The starting condition for the current ensemble, 'L'eft or
        'R'ight.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        :py:const:`.path._STATUS`.

    """
    path_old = ensemble['path_ensemble'].last_path
    intf = ensemble['interfaces']
    ph_pt1, ph_pt2 = crossing_finder(path_old, intf[1])
    if ph_pt1 is None and ph_pt2 is None:
        return False, path_old, 'NCR'
    sub_ens = {'interfaces': [intf[1], intf[1], intf[2]],
               'order_function': ensemble['order_function']}
    osc_try = 0  # One step crossing attempt counter
    success = False
    for i in range(tis_settings['n_jumps']):
        logger.debug('Trying a new stone skipping move, jump %i', i)
        # Here we choose between the two
        # possible shooting points that describe a crossing.
        sh_pt = ph_pt1 if ensemble['rgen'].rand() >= 0.5 else ph_pt2
        ensemble['engine'].dump_phasepoint(sh_pt, str(counter()) + '_ss_shoot')
        # To continue, we must be sure that the new path
        # CROSSES the interface in ONLY ONE step.
        # Generate paths until it succeed. That is
        # what makes this version of the SS move useless for large systems.
        for j in range(tis_settings['maxlength']):
            # This function can become actually fun to work on.
            # e.g. have a 50% chance to give random v for each particle
            # Modify the velocities:
            # todo modify_v could just use system directly
            logger.debug("jump %s, try %s, start: %s", i, j, sh_pt.order[0])
            ensemble['system'] = sh_pt.copy()
            ensemble['engine'].modify_velocities(ensemble, tis_settings)
            # A path of two frames is going to be generated.
            success, path = one_step_crossing(ensemble, intf[1])
            osc_try += 1
            if osc_try > 5 * tis_settings['n_jumps'] and \
                    path_old.get_move() in {'ld', 'ki', 'is'}:
                logger.info('Performing a shooting move before the use of ss')
                success, trial_path, status = shoot(ensemble,
                                                    tis_settings,
                                                    start_cond, sh_pt)
                trial_path.set_move('is')
                return success, trial_path, status

            if success:
                break
        else:  # In case we reached maxlength in jumps attempts.
            success = False
            path.status = 'NSS'
            trial_path = path
            break

        # Depending on the shooting point (before or after the interface),
        # a backward path or a continuation has to be generated.
        new_segment = path.empty_path(maxlen=tis_settings['maxlength'] - 1)
        if path.get_end_point(intf[1], intf[2]) == start_cond:
            path = path.reverse(ensemble['order_function'])
        sub_ens['system'] = path.phasepoints[1].copy()
        success, _ = ensemble['engine'].propagate(new_segment, sub_ens)
        new_segment.phasepoints.insert(0, path.phasepoints[0].copy())

        if not success:
            new_segment.status = 'XSS'
            trial_path = new_segment
            break

        ph_pt1, ph_pt2 = crossing_finder(new_segment, intf[1], last_frame=True)

    logger.debug('SS web: %s, one step crossing tries: %s', success, osc_try)

    if success:
        if ensemble['rgen'].rand() < 0.5:
            new_segment = new_segment.reverse(ensemble['order_function'])
        success, trial_path, _ = extender(new_segment, ensemble,
                                          tis_settings, start_cond)

    if success:
        success, trial_path = ss_wt_wf_acceptance(trial_path, ensemble,
                                                  tis_settings,
                                                  start_cond)

    trial_path.generated = ('ss', sh_pt.order[0], osc_try, trial_path.length)

    logger.debug('SS move: %s', trial_path.status)
    if not success:
        return False, trial_path, trial_path.status

    # This might get triggered when accepting 0-L paths.
    assert start_cond == trial_path.get_start_point(intf[0], intf[2]), \
        'SS: Path has an implausible start.'

    trial_path.status = 'ACC'

    return True, trial_path, trial_path.status


def ss_wt_wf_metropolis_acc(path_new, ensemble,
                            tis_settings, start_cond='L'):
    """Accept or reject the path_new.

    Super detailed balance rule is used in the original version
    and in the High Acceptance one for SS and WF.

    In the regular version, P acc = min (1, Cold/Cnew), where
    for Stone Skipping C is crossing, for Web Throwing C is segment,
    for Wire Fencing C is number of phasepoint between ensemble and
    right interface.

    In the High Acceptance version, P acc = 1 for SS and Wf. It
    also allows to accept paths that go from B to A, by reversing them.
    NB. To respect super detailed balance, the weights have to be changed
    accordingly. This is done elsewhere.

    Parameters
    ----------
    path_new : object like :py:class:`.PathBase`
        This is the new path that might get accepted.
    ensemble : dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces` : list/tuple of floats
          These are the interface positions of the form
          ``[left, middle, right]``.

    tis_settings : dict
        This contains the settings for TIS. Keys used here:

        * `high_accept` : boolean, the option for High Acceptance SS/WF.

    start_cond : string, optional
        The starting condition for the current ensemble, 'L'eft or
        'R'ight.

    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.

    """
    interfaces = ensemble['interfaces']
    path_old = ensemble['path_ensemble'].last_path
    if tis_settings.get('shooting_move') == 'wt':
        sour_int = tis_settings['interface_sour']
        cr_old = segments_counter(path_old, sour_int, interfaces[1])
        cr_new = segments_counter(path_new, sour_int, interfaces[1])
        if ensemble['rgen'].rand() >= min(1.0, cr_old / cr_new):
            path_new.status = 'WTA'
            return False

    else:
        if not tis_settings.get('high_accept', False):
            if tis_settings.get('shooting_move') == 'ss':
                cr_old = crossing_counter(path_old, interfaces[1])
                cr_new = crossing_counter(path_new, interfaces[1])
                if ensemble['rgen'].rand() >= min(1.0, cr_old / cr_new):
                    path_new.status = 'SSA'
                    return False
            elif tis_settings.get('shooting_move') == 'wf':
                wf_cap = tis_settings.get('interface_cap', interfaces[2])
                cr_old, _ = wirefence_weight_and_pick(path_old, interfaces[1],
                                                      wf_cap)
                cr_new, _ = wirefence_weight_and_pick(path_new, interfaces[1],
                                                      wf_cap)
                if ensemble['rgen'].rand() >= min(1.0, cr_old / cr_new):
                    path_new.status = 'WFA'
                    return False

    if start_cond != path_new.get_start_point(interfaces[0], interfaces[2]):
        path_new.status = 'BWI'
        return False
    path_new.status = 'ACC'
    return True


def web_throwing(ensemble, tis_set, start_cond='L'):
    """Perform a web_throwing move.

    This function performs the great web throwing move from an initial path.

    Parameters
    ----------
    ensemble : dict
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the TIS step for.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is the random generator that will be used.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_set : dict
        This contains the settings for TIS. Keys used here:

        * `aimless`: boolean, is the shooting aimless or not?
        * `allowmaxlength`: boolean, should paths be allowed to reach
          maximum length?
        * `maxlength`: integer, maximum allowed length of paths.

    start_cond : string, optional
        The starting condition for the current ensemble, 'L'eft or
        'R'ight.


    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        :py:const:`.path._STATUS`.

    """
    path_old = ensemble['path_ensemble'].last_path
    interfaces = ensemble['interfaces']
    sour = tis_set['interface_sour']
    assert interfaces[0] < sour <= interfaces[1], \
        'SOUR interface is not correctly positioned'

    ccnt = segments_counter(path_old, sour, interfaces[1])
    if ccnt == 0:
        return False, path_old, 'NSG'

    seg_i = int(ensemble['rgen'].rand()[0] * ccnt)
    wt_int = [sour, sour, ensemble['interfaces'][1]]
    source_seg = select_and_trim_a_segment(path_old, sour, wt_int[2], seg_i)
    sub_ens = {'interfaces': wt_int,
               'order_function': ensemble['order_function']}

    shoots, save_acc = [0], 0

    key = ensemble['rgen'].rand() >= 0.5  # Start from a random side
    for _ in range(tis_set['n_jumps']):
        if ensemble['rgen'].rand() >= 0.5:
            shoots[-1] += 1  # One more on the Same side
        else:
            shoots.append(1)  # A move in the other side

    for n_virtual in shoots:
        key = not key  # Change side, key controls also path reverse
        for _ in range(n_virtual):
            if key:
                pre_shooting_point = source_seg.phasepoints[-1]
                shooting_point = source_seg.phasepoints[-2]
            else:
                pre_shooting_point = source_seg.phasepoints[0]
                shooting_point = source_seg.phasepoints[1]

            prefix = str(counter())
            ensemble['engine'].dump_phasepoint(pre_shooting_point,
                                               prefix + '_wt_pre_shoot')
            ensemble['engine'].dump_phasepoint(shooting_point,
                                               prefix + '_wt_shoot')

            new_seg = path_old.empty_path(maxlen=tis_set['maxlength'])
            new_seg.append(pre_shooting_point)
            logger.debug('Trying a new web')
            sub_ens['system'] = shooting_point.copy()
            ensemble['engine'].propagate(new_seg, sub_ens, reverse=key)
            start = new_seg.get_start_point(wt_int[0], wt_int[-1])
            end = new_seg.get_end_point(wt_int[0], wt_int[-1])
            logger.debug('WT web starts %s, ends %s, reverse %s',
                         start, end, key)
            if segments_counter(new_seg, sour, wt_int[2], reverse=key) == 1:
                logger.debug('Web successful')
                source_seg = new_seg.reverse(ensemble['order_function'],
                                             rev_v=False) if key else new_seg
                source_seg.status = 'ACC'
                save_acc += 1
                break

    logger.debug('WT segments accepted: %s', save_acc)

    accept, trial_path, _ = extender(source_seg, ensemble, tis_set, start_cond)

    trial_path.generated = ('wt', source_seg.phasepoints[1].order[0],
                            save_acc, trial_path.length)
    # Also Check that we did not get a B to A or a B to B path.
    if accept:
        accept, trial_path = ss_wt_wf_acceptance(trial_path, ensemble, tis_set)
    logger.debug('WT move: %s', trial_path.status)

    # Set the path flags
    if accept and path_old.get_move() == 'ld' and save_acc == 0:
        trial_path.set_move('ld')

    return accept, trial_path, trial_path.status


def extender(source_seg, ensemble, tis_set, start_cond=('R', 'L')):
    """Extend a path to the given interfaces.

    This function will perform the web throwing move from an initial path.

    Parameters
    ----------
    source_seg : object like :py:class:`.PathBase`
        This is the input path which will be prolonged.
    ensemble : dictionary of objects
        It contains:

        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `interfaces`: list of floats
          These are the interface positions on form
          [left, middle, right].

    tis_set : dict
        This contains the settings for TIS. Keys used here:

        * `aimless`: boolean, is the shooting aimless or not?
        * `allowmaxlength`: boolean, should paths be allowed to reach
          maximum length?
        * `maxlength`: integer, maximum allowed length of paths.
        * `high_accept`: boolean, the option for High Acceptance WF and SS.

    start_cond : string or tuple of strings, optional
        The starting condition for the current ensemble, 'L'eft or
        'R'ight. ('L', 'R'), the default option, implies no directional
        difference.


    Returns
    -------
    out[0] : boolean
        True if the path can be accepted.
    out[1] : object like :py:class:`.PathBase`
        Returns the generated path.
    out[2] : string
        Status of the path, this is one of the strings defined in
        :py:const:`.path._STATUS`.

    """
    interfaces = ensemble['interfaces']
    ensemble['system'] = source_seg.phasepoints[0].copy()

    # Extender
    if interfaces[0] <= ensemble['system'].order[0] < interfaces[-1]:
        back_segment = source_seg.empty_path(maxlen=tis_set['maxlength'])
        logger.debug('Trying to extend backwards')
        source_seg_copy = source_seg.copy()

        if not shoot_backwards(back_segment, source_seg_copy, ensemble,
                               tis_set, start_cond):
            if not tis_set.get('high_accept', False):
                return False, source_seg_copy, source_seg_copy.status

        trial_path = paste_paths(back_segment, source_seg, overlap=True,
                                 maxlen=tis_set['maxlength'])
    else:
        trial_path = source_seg.copy()

    ensemble['system'] = trial_path.phasepoints[-1].copy()
    if interfaces[0] <= ensemble['system'].order[0] < interfaces[-1]:
        forth_segment = source_seg.empty_path(maxlen=tis_set['maxlength'])
        ensemble['engine'].propagate(forth_segment, ensemble)

        trial_path.phasepoints = trial_path.phasepoints[:-1] + \
            forth_segment.phasepoints

    if trial_path.length >= tis_set['maxlength']:
        trial_path.status = 'FTX'  # exceeds "memory".
        return False, trial_path, trial_path.status
    trial_path.status = 'ACC'
    return True, trial_path, trial_path.status



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Plot raw data from a simulation."""
# pylint: disable=invalid-name
import os
import sys
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.cm import get_cmap
from pyretis.core import create_box, System, Particles
from pyretis.setup import create_force_field, create_orderparameter
from pyretis.inout.settings import parse_settings_file
from pyretis.inout.formats.path import PathIntFile


def plot_potential(settings, axi, axj):
    """Plot the potential in the given axis."""
    forcefield = create_force_field(settings)
    box = create_box(periodic=[False, False])
    fakesys = System(units='reduced', box=box)
    fakesys.particles = Particles(dim=2)
    fakesys.add_particle(name='B', pos=np.zeros(2), ptype=1)
    minx, maxx = -0.5, 0.5
    miny, maxy = -1.0, 1.0
    xval = np.linspace(minx, maxx, 100)
    yval = np.linspace(miny, maxy, 100)
    xpos, ypos = np.meshgrid(xval, yval, indexing='ij')
    pot = np.zeros_like(xpos)
    for i, x in enumerate(xval):
        for j, y in enumerate(yval):
            fakesys.particles.pos[0, 0] = x
            fakesys.particles.pos[0, 1] = y
            pot[i, j] = forcefield.evaluate_potential(fakesys)
    axi.contourf(xpos, ypos, pot, 10, cmap=get_cmap('viridis'), alpha=0.8)
    # add interfaces
    for inter in settings['simulation']['interfaces']:
        axi.axvline(x=inter, lw=2, ls=':', color='#262626', alpha=0.8)
        axj.axhline(y=inter, lw=2, ls=':', color='#262626', alpha=0.8)
    extra_int = [settings['orderparameter']['inter_a'],
                 settings['orderparameter']['inter_b']]
    for inter in extra_int:
        axi.axvline(x=inter, lw=2, ls=':', color='#262626', alpha=0.5)
        axj.axhline(y=inter, lw=2, ls=':', color='#262626', alpha=0.5)
    axi.set_xlim((minx, maxx))
    axi.set_ylim((miny, maxy))
    axi.set_xlabel((r'Position ($x$)'), fontsize='large')
    axi.set_ylabel((r'Position ($y$)'), fontsize='large')
    axj.set_xlabel((r'Step number'), fontsize='large')
    axj.set_ylabel((r'Order parameter ($\lambda$)'), fontsize='large')


def plot_ensemble(settings, dirname, axi, axj, maxlines=100, minorder=None,
                  skip=1):
    """Plot trajectories from an ensemble."""
    orderp = create_orderparameter(settings)
    box = create_box(periodic=[False, False])
    fakesys = System(units='reduced', box=box)
    fakesys.particles = Particles(dim=2)
    fakesys.add_particle(name='B', pos=np.zeros(2), ptype=1)
    forcefield = create_force_field(settings)
    traj_file = os.path.join(dirname, 'traj.txt')
    iplot = 0
    all_lines = []
    all_lines2 = []
    last_point = []
    first_point = []
    order_last = []
    order_last2 = []
    with PathIntFile(traj_file, 'r') as tfile:
        for i, traj in enumerate(tfile.load()):
            if traj['comment'][0].split('status:')[-1].strip() != 'ACC':
                continue
            if i % skip != 0:
                continue
            pos = np.array([x['pos'][0] for x in traj['data']])
            order = []
            for posi in pos:
                fakesys.particles.pos[0, 0] = posi[0]
                fakesys.particles.pos[0, 1] = posi[1]
                fakesys.particles.vpot = forcefield.evaluate_potential(fakesys)
                order.append(orderp.calculate(fakesys)[0])
            if minorder is not None:
                if max(order) < minorder:
                    continue
            line, = axi.plot(pos[:, 0], pos[:, 1], lw=3, alpha=0.9)
            line2, = axj.plot(order, lw=3, alpha=0.9)
            order_last.append((len(order) - 1, order[-1]))
            order_last2.append((len(order) - 2, order[-2]))
            all_lines.append(line)
            all_lines2.append(line2)
            first_point.append((pos[0, 0], pos[0, 1]))
            last_point.append((pos[-1, 0], pos[-1, 1]))
            iplot += 1
            if iplot >= maxlines:
                break
    # Add colors now that we know how many we have created:
    cmap = get_cmap(name='coolwarm')
    colors = cmap(np.linspace(0, 1, iplot))
    for i, (line, line2) in enumerate(zip(all_lines, all_lines2)):
        line.set_color(colors[i])
        axi.scatter(first_point[i][0], first_point[i][1], s=50, marker='x',
                    color=line.get_color(), alpha=0.9)
        axi.scatter(last_point[i][0], last_point[i][1], s=50, marker='o',
                    color=line.get_color(), alpha=0.9)
        line2.set_color(colors[i])
        if order_last2[i][1] < order_last[i][1]:
            end = '^'
        else:
            end = 'v'
        axj.scatter(order_last[i][0], order_last[i][1], s=50, marker=end,
                    color=line2.get_color(), alpha=0.9)
        axj.scatter(order_last2[i][0], order_last2[i][1], s=50, marker='x',
                    color=line2.get_color(), alpha=0.9)


if __name__ == '__main__':
    ens = sys.argv[1]
    sim_settings = parse_settings_file('retis.rst')
    fig = plt.figure(figsize=(12, 6))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    plot_potential(sim_settings, ax1, ax2)
    plot_ensemble(sim_settings, ens, ax1, ax2, maxlines=25, skip=2,
                  minorder=None)
    plt.subplots_adjust(right=0.95, left=0.10, top=0.95,
                        bottom=0.10, wspace=0.3)
    plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Double pendulum example."""
# pylint: disable=invalid-name
import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib import animation
from matplotlib import gridspec
from pyretis.core.units import create_conversion_factors
from pyretis.setup import (create_system, create_engine,
                           create_force_field, create_simulation)

# Define potential parameters:
L1 = 1.5
L2 = 2.0
M1 = 1.0
M2 = 1.0
L1SQ = L1**2
L2SQ = L2**2
THETA1 = 45. * np.pi / 180.
THETA2 = 90. * np.pi / 180.
DTHETA1 = 0.0
DTHETA2 = 0.0
STEPS_PER_FRAME = 10
# Give simulation settings:
settings = {}
settings['simulation'] = {
    'task': 'md-nve',
    'steps': 100000,
    'exe_path': None
}
settings['system'] = {
    'units': 'gromacs',
    'temperature': 1.,
    'dimensions': 2
}
settings['engine'] = {
    'class': 'VVIntegrator',
    'timestep': 0.005,
    'module': 'myintegrator.py'
}
settings['output'] = {
    'backup': 'overwrite',
    'energy-file': 1,
    'screen': 10,
    'trajectory-file': 10
}
settings['potential'] = [
    {'class': 'DoublePendulum',
     'module': 'doublependulum.py',
     'g': 1.0, 'l1': L1, 'l2': L2}
]
settings['particles'] = {
    'position': {'generate': 'sq', 'repeat': [2, 1], 'lcon': 1},
    'ptype': [0, 0],
    'name': ['A', 'B'],
    'mass': {'A': M1, 'B': M2}
}

UNIT = settings['system']['units']
create_conversion_factors(UNIT)
print('# Creating system from settings.')
system = create_system(settings)
system.forcefield = create_force_field(settings)
system.particles.pos[0][0] = THETA1
system.particles.pos[1][0] = THETA2
system.particles.vel[0][0] = DTHETA1
system.particles.vel[1][0] = DTHETA2
settings['system']['obj'] = system
simulation = create_simulation(settings)
mpl.rc('axes', labelsize='large')
mpl.rc('font', family='serif')
fig = plt.figure(figsize=(12, 6))
grid = gridspec.GridSpec(3, 2)
ax1 = fig.add_subplot(grid[:, 0])
ax1.set_xlim((-5, 5))
ax1.set_ylim((-5, 2))
ax1.set_aspect('equal')
time_text = ax1.text(0.02, 0.90, '', transform=ax1.transAxes)

# set up circles to represent the two objects:
circ1 = plt.Circle((0, 0), radius=0.25, alpha=0.6)
circ1.set_visible(False)
ax1.add_patch(circ1)
circ2 = plt.Circle((0, 0), radius=0.25, alpha=0.6)
circ2.set_visible(False)
ax1.add_patch(circ2)
# Add one for the orrigin
circ0 = plt.Circle((0, 0), radius=0.1, alpha=0.6, color='#262626')
ax1.add_patch(circ0)
# Add a lines for the connection between them
line1, = ax1.plot([], [], lw=3, ls='-', color='#262626', alpha=0.5)
line2, = ax1.plot([], [], lw=3, ls='-', color='#262626', alpha=0.5)
# add second axis for displaying energies:
ax2 = fig.add_subplot(grid[0, 1])
linepot, = ax2.plot([], [], lw=3, ls='-', alpha=0.8)
ax2.set_ylim(-7.5, 0)
ax3 = fig.add_subplot(grid[1, 1])
linekin, = ax3.plot([], [], lw=3, ls='-', alpha=0.8)
ax3.set_ylim(0, 3.5)
ax4 = fig.add_subplot(grid[2, 1])
linetot, = ax4.plot([], [], lw=3, ls='-', alpha=0.8)
ax4.set_ylim(-2.5, -0.5)
ax4.set_xlabel('Steps')
ax2.set_xticks([])
ax3.set_xticks([])


def init():
    """Declare what to re-draw when clearing the animation frame.

    Returns
    -------
    out : list
        List of the patches to be drawn.

    """
    patches = []
    circ1.set_visible(False)
    circ2.set_visible(False)
    patches.append(circ1)
    patches.append(circ2)
    line1.set_data([], [])
    patches.append(line1)
    line2.set_data([], [])
    patches.append(line2)
    time_text.set_text('')
    patches.append(time_text)
    return patches


def update(frame, sys, sim):
    """Update plots for the animation.

    This function will be running the simulation and updating the plots.
    It is called one time per step, and we choose to update the simulation
    inside this function

    Parameters
    ----------
    frame : int
        The current frame number, supplied by `animation.FuncAnimation`.
    sys : object like :py:class:`.System`
        The system object we are simulating.
    sim : object like :py:class:`.Simulation`
        The simulation we are running.

    Returns
    -------
    out : list
        List of the patches to be drawn.

    """
    patches = []
    if not sim.is_finished():
        for _ in range(STEPS_PER_FRAME):
            result = sim.step()
        pos = sys.particles.pos
        theta1 = pos[0][0]
        costheta1 = np.cos(theta1)
        sintheta1 = np.sin(theta1)
        x1, y1 = L1 * sintheta1, -L1 * costheta1
        circ1.center = np.array([x1, y1])
        circ1.set_visible(True)
        patches.append(circ1)
        theta2 = pos[1][0]
        costheta2 = np.cos(theta2)
        sintheta2 = np.sin(theta2)
        x2, y2 = x1 + L2*sintheta2, y1 - L2 * costheta2
        circ2.center = np.array([x2, y2])
        circ2.set_visible(True)
        patches.append(circ2)
        line1.set_data([0.0, x1], [0.0, y1])
        patches.append(line1)
        line2.set_data([x1, x2], [y1, y2])
        patches.append(line2)
        vpot = result['thermo']['vpot']
        vel = sys.particles.vel
        dtheta1 = vel[0][0]
        dtheta2 = vel[1][0]
        dt1sq = dtheta1**2
        dt2sq = dtheta2**2
        ekin = (0.5*(M1+M2)*L1SQ*dt1sq + 0.5*M2*L2SQ*dt2sq +
                M2*L1*L2*dtheta1*dtheta2*np.cos(theta1-theta2))
        vpot = 2.0 * result['thermo']['vpot']
        etot = ekin + vpot
        linekin.set_xdata(np.append(linekin.get_xdata(), frame))
        linekin.set_ydata(np.append(linekin.get_ydata(), ekin))
        patches.append(linekin)
        linepot.set_xdata(np.append(linepot.get_xdata(), frame))
        linepot.set_ydata(np.append(linepot.get_ydata(), vpot))
        patches.append(linepot)
        linetot.set_xdata(np.append(linetot.get_xdata(), frame))
        linetot.set_ydata(np.append(linetot.get_ydata(), etot))
        patches.append(linetot)
        ax2.set_xlim(0, frame + 1)
        ax3.set_xlim(0, frame + 1)
        ax4.set_xlim(0, frame + 1)
        time_text.set_text(f"Step: {result['cycle']['step']}")
        patches.append(time_text)
        return patches
    print('Simulation is done')
    return patches


# This will run the animation/simulation:
FMAX = int(float(settings['simulation']['steps']) / float(STEPS_PER_FRAME))
anim = animation.FuncAnimation(fig, update,
                               frames=FMAX+1,
                               fargs=[system, simulation],
                               repeat=False, interval=2, blit=True,
                               init_func=init)
# for making a movie:
# anim.save('particles.mp4', fps=30, extra_args=['-vcodec', 'libx264'])
plt.show()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods that will output results from the analysis functions.

The Methods defined here will also run the analysis and output
according to given settings.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

analyse_file (:py:func:`.analyse_file`)
    Method to analyse a file. For example, it can be used as

    >>> from pyretis.inout.analysisio import analyse_file
    >>> result, raw_data = analyse_file('cross', 'cross.txt', settings)

    To output the results, for instance by plotting, one can do:

    >>> figures = output_results(file_type, plotter, result, raw_data)
    >>> txt_file = output_results(file_type, txt_plotter, result, raw_data)


run_analysis_files (:py:func:`.run_analysis_files`)
    Methods to the analysis on a set of files. It will create output
    that can be used for reporting.
"""
import logging
import os
import numpy as np
from pyretis.core.units import create_conversion_factors
from pyretis.core.pathensemble import generate_ensemble_name
from pyretis.analysis import (analyse_flux, analyse_energies, analyse_orderp,
                              analyse_path_ensemble, analyse_repptis_ensemble,
                              match_probabilities, retis_flux, retis_rate,
                              perm_calculations)
from pyretis.inout import print_to_screen
from pyretis.inout.formats.formatter import format_number
from pyretis.inout.plotting import create_plotter
from pyretis.inout.plotting import TxtPlotter
from pyretis.inout.report import generate_report
from pyretis.inout.settings import SECTIONS, copy_settings
from pyretis.inout.formats.pathensemble import PathEnsembleFile
from pyretis.inout.formats.energy import EnergyFile
from pyretis.inout.formats.order import OrderFile
from pyretis.inout.formats.cross import CrossFile
from pyretis.inout.simulationio import OUTPUT_TASKS
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = ['analyse_file', 'run_analysis_files',
           'run_analysis', 'read_first_block',
           'repptis_running_pcross_analysis']


_ANALYSIS_FUNCTIONS = {'cross': analyse_flux,
                       'order': analyse_orderp,
                       'energy': analyse_energies,
                       'pathensemble': analyse_path_ensemble,
                       'pathensemble_repptis': analyse_repptis_ensemble}


# Input files for analysis
_FILES = {'md-flux': {},
          'md-nve': {},
          'md': {},
          'tis': {},
          'make-tis-files': {},
          'retis': {},
          'repptis': {}}
# Add files
for key in ('cross', 'energy', 'order'):
    _FILES['md-flux'][key] = OUTPUT_TASKS[key]['filename']
for key in ('energy',):
    _FILES['md-nve'][key] = OUTPUT_TASKS[key]['filename']
    _FILES['md'][key] = OUTPUT_TASKS[key]['filename']
for key in ('pathensemble',):
    for key2 in ('make-tis-files', 'tis', 'retis', 'repptis'):
        _FILES[key2][key] = OUTPUT_TASKS[key]['filename']


def run_analysis(settings):
    """Run a predefined analysis task.

    Parameters
    ----------
    settings : dict
        Simulation settings and settings for the analysis.

    Returns
    -------
    out : dict
        A dictionary with the results from the analysis. This dict
        can be used to generate a report.

    """
    runners = {'retis': run_retis_analysis,
               'tis': run_tis_analysis,
               'make-tis-files': run_tis_analysis,
               'md-flux': run_mdflux_analysis,
               'repptis': run_repptis_analysis}
    sim_task = settings['simulation']['task'].lower()
    report_dir = settings['analysis'].get('report-dir', None)
    plotter = create_plotter(settings['analysis']['plot'],
                             out_dir=report_dir)
    backup = settings['output'].get('backup', SECTIONS['output']['backup'])
    txt_plotter = TxtPlotter(settings['analysis']['txt-output'],
                             backup=backup,
                             out_dir=report_dir)
    if sim_task in runners.keys():
        runner = runners[sim_task]
        return runner(settings, plotter, txt_plotter)
    msgtxt = f'Unknown analysis task "{sim_task}" requested!'
    logger.error(msgtxt)
    raise ValueError(msgtxt)


def get_path_ensemble_files(ensemble_number, settings, detect,
                            interfaces):
    """Return files for a single path ensemble.

    Here, we will return the files needed to analyse a single path
    ensemble and we will also return settings which can be used for
    the analysis.

    Parameters
    ----------
    ensemble_number : int
        This is the integer representing the ensemble.
    settings : dict
        The settings to use for an analysis/simulation
    detect : float or None
        The interface use for detecting if a path is successful for not.
    interfaces : list of floats
        The interfaces used for this particular path simulation.

    Returns
    -------
    local_settings : dict
        This dict contains settings which can be used for a initial
        flux analysis.
    files : list of tuples
        The tuples in this list are the files which can be analysed
        further, using the settings in `out[0]`.

    """
    sim_task = settings['simulation']['task'].lower()
    lsetting = copy_settings(settings)
    lsetting['simulation']['interfaces'] = interfaces
    lsetting['tis']['detect'] = detect
    lsetting['tis']['ensemble_number'] = ensemble_number
    run_path = settings['simulation']['exe-path']
    files = []
    for file_type in _FILES[sim_task]:
        filename = os.path.join(run_path,
                                generate_ensemble_name(ensemble_number),
                                _FILES[sim_task][file_type])
        if os.path.isfile(filename):
            files.append((file_type, filename))
            logger.debug('Adding file "%s" for analysis', filename)
        else:
            logger.debug('Could not find file %s', filename)
    return lsetting, files


def get_path_simulation_files(sim_settings):
    """Set up for analysis of TIS and RETIS simulations.

    For these kinds of simulations, we expect to analyse several
    directories with raw-data. For TIS we expect to find a directory
    with raw-data for the initial flux (named ``flux``) and the
    directories for the path simulations (named ``001``, ``002`` etc.
    for ensembles [0^+], [1^+] and so on). For RETIS, we expect to
    find the same directories but with a ``000`` (for the ``[0^-]``
    ensemble) rather than the ``flux`` directory.

    Parameters
    ----------
    sim_settings : dict
        The settings used for the (RE)TIS simulation(s). These settings
        are used to get the interfaces used and the path ensembles
        defined by these interfaces.

    Returns
    -------
    all_settings : list of dict
        This dict in this list contains settings which can be used for
        the analysis.
    all_files : list of lists of tuples
        `all_files[i]` is a list of files from path ensemble
        simulation `i`. For TIS, `all_files[0]` should be the files
        obtained in the initial flux simulation. These files can be
        analysed using the settings in `all_settings[i]`.

    """
    # Check if we can do flux analysis:
    all_files, all_settings = [], []
    interfaces = sim_settings['simulation']['interfaces']
    reactant = interfaces[0]
    product = interfaces[-1]
    task = sim_settings['simulation']['task']
    # for PPTIS or REPPTIS: memory
    mem = sim_settings.get(task, {}).get('memory', 1)

    if task == 'make-tis-files':
        all_files.append(None)
        all_settings.append(None)
    else:  # just add the 0 ensemble
        detect = None
        interface = [-float('inf'), reactant, reactant]
        setts, files = get_path_ensemble_files(0, sim_settings, detect,
                                               interface)
        all_files.append(files)
        all_settings.append(setts)
    for i, middle in enumerate(interfaces[:-1]):
        detect = interfaces[i + 1] if i < len(interfaces) else product
        if task in ['pptis', 'repptis']:
            reactant = interfaces[max(0, i - mem)]
            product = interfaces[min(i + mem, len(interfaces) - 1)]
            interface = [reactant, middle, product]
        else:
            interface = [reactant, middle, product]
        setts, files = get_path_ensemble_files(i + 1, sim_settings, detect,
                                               interface)
        all_files.append(files)
        all_settings.append(setts)
    return all_settings, all_files


def print_value_error(heading, value, rel_error, level=None):
    """Print out the matched probabilities."""
    val = format_number(value, 0.1, 100)
    msgtxt = f'{heading}: {val}'
    print_to_screen(msgtxt.strip(), level=level)
    fmt_scale = format_number(rel_error * 100, 0.1, 100)
    msgtxt = f'(Relative error: {fmt_scale.rstrip()} %)'
    print_to_screen(msgtxt, level=level)


def run_single_tis_analysis(settings, plotter, txt_plotter):
    """Run the analysis for a single TIS ensemble.

    Parameters
    ----------
    settings : dict
        The settings to use for an analysis/simulation.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    Returns
    -------
    out : list or dict or similar
        The output from the analysis.

    """
    sim = settings['simulation']
    tis = settings['ensemble'][0]['tis']
    sett, files = get_path_ensemble_files(
        tis['ensemble_number'], settings,
        tis.get('detect', sim['interfaces'][-1]),
        sim['interfaces'])
    msgtxt = f"Analysing ensemble {tis['ensemble_number']}"
    print_to_screen(msgtxt, level='info')
    print_to_screen()
    result = run_analysis_files(sett, files, plotter, txt_plotter)
    report_txt = generate_report('tis', result, output='txt')[0]
    print_to_screen(''.join(report_txt))
    print_to_screen()
    return result


def run_tis_analysis(settings, plotter, txt_plotter):
    """Run the analysis for TIS.

    Parameters
    ----------
    settings : dict
        The settings to use for an analysis/simulation.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    Returns
    -------
    out : list or dict or similar
        The output from the analysis.

    """
    list_ens = settings['ensemble']
    if len(list_ens) == 1 and settings['simulation']['task'] == 'tis':
        return run_single_tis_analysis(settings, plotter, txt_plotter)
    results = {'cross': None, 'pathensemble': [], 'matched': None}
    all_settings, all_files = get_path_simulation_files(settings)
    nens = len(all_settings) - 1
    for i, (sett, files) in enumerate(zip(all_settings, all_files)):
        if i == 0:
            msgtxt = ('Initial flux is not calculated here.\n'
                      'Remember to calculate this separately!')
            logger.info(msgtxt)
            print_to_screen(msgtxt, level='warning')
        else:
            msgtxt = f'Analysing ensemble {i} of {nens}'
            print_to_screen(msgtxt, level='info')
            print_to_screen()
            result = run_analysis_files(sett, files, plotter, txt_plotter)
            results['pathensemble'].append(result['pathensemble'])
            report_txt = generate_report('tis', result,
                                         output='txt')[0]
            print_to_screen(''.join(report_txt))
            print_to_screen()
    # match probabilities:
    out, fig, txt = analyse_and_output_matched(results['pathensemble'],
                                               plotter, txt_plotter,
                                               settings)
    results['matched'] = {'out': out, 'figures': fig, 'txtfile': txt}
    print_to_screen('Overall results', level='success')
    print_to_screen('===============', level='success')
    print_to_screen()
    print_value_error('TIS Crossing probability',
                      out['prob'], out['relerror'], level='success')
    return results


def run_retis_analysis(settings, plotter, txt_plotter):
    """Run the analysis for RETIS.

    Parameters
    ----------
    settings : dict
        The settings to use for an analysis/simulation.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    """
    units = settings['system']['units']
    permeability = settings['simulation'].get('permeability', False)
    create_conversion_factors(units, **settings.get('unit-system', {}))
    all_settings, all_files = get_path_simulation_files(settings)
    results = {'cross': None,
               'pathensemble': [],
               'matched': None,
               'permeability': permeability}
    nens = len(all_settings) - 1
    print_to_screen()
    for i, (sett, files) in enumerate(zip(all_settings, all_files)):
        msgtxt = f'Analysing ensemble {i} of {nens}'
        print_to_screen(msgtxt, level='info')
        print_to_screen()
        if i == 0:
            result = run_analysis_files(sett, files, plotter, txt_plotter)
            results['pathensemble0'] = result['pathensemble']
            report_txt = generate_report('retis0', result, output='txt')[0]
            print_to_screen(''.join(report_txt))
            print_to_screen()
        else:
            result = run_analysis_files(sett, files, plotter, txt_plotter)
            results['pathensemble'].append(result['pathensemble'])
            report_txt = generate_report('tis', result,
                                         output='txt')[0]
            print_to_screen(''.join(report_txt))
            print_to_screen()
    # flux first:
    time_subcycles = settings['engine'].get('subcycles', 1)
    timestep = settings['engine']['timestep'] * time_subcycles
    flux, flux_error = retis_flux(results['pathensemble0']['out'],
                                  results['pathensemble'][0]['out'],
                                  timestep)
    results['flux'] = {'value': flux, 'error': flux_error,
                       'unit': units}
    # match probabilities:
    out, fig, txt = analyse_and_output_matched(results['pathensemble'],
                                               plotter, txt_plotter,
                                               settings,
                                               flux=flux)
    results['matched'] = {'out': out, 'figures': fig, 'txtfile': txt}
    rate, rate_error = retis_rate(out['prob'], out['relerror'],
                                  flux, flux_error)
    results['rate'] = {'value': rate, 'error': rate_error,
                       'unit': units}

    xi, tau, perm, xi_err, tau_err, perm_err = perm_calculations(
        results['pathensemble0'], out['prob'], out['relerror'])
    xi_figs, tau_figs = perm_figures(results['pathensemble0'], plotter)
    results['xi'] = {'value': xi,
                     'error': xi_err,
                     'fig': xi_figs}
    results['tau'] = {'value': tau,
                      'error': tau_err,
                      'fig': tau_figs}
    results['perm'] = {'value': perm,
                       'error': perm_err}

    print_to_screen('Overall results', level='success')
    print_to_screen('===============', level='success')
    print_to_screen()
    print_value_error('RETIS Crossing probability',
                      out['prob'], out['relerror'], level='success')
    print_to_screen()
    print_value_error(f'Initial flux (units 1/{units})', flux,
                      flux_error, level='success')
    print_to_screen()
    print_value_error(f'Rate constant (units 1/{units})', rate,
                      rate_error, level='success')
    if permeability:
        print_to_screen()
        print_value_error('Xi', xi, xi_err, level='success')
        print_to_screen()
        print_value_error(f'Tau/dz (unit: {units}/OP-unit)',
                          tau, tau_err, level='success')
        print_to_screen()
        print_value_error(f'Permeability (unit: OP-unit/{units})',
                          perm, perm_err, level='success')
    return results


def run_repptis_analysis(settings, plotter, txt_plotter):
    """Run the analysis for PPRETIS.

    Parameters
    ----------
    settings : dict
        The settings to use for an analysis/simulation.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    """
    units = settings['system']['units']
    permeability = settings['simulation'].get('permeability', False)
    create_conversion_factors(units, **settings.get('unit-system', {}))
    all_settings, all_files = get_path_simulation_files(settings)
    results = {'cross': None,
               'pathensemble': [],
               'matched': None,
               'permeability': permeability,
               'reptis': {}}
    nens = len(all_settings) - 1
    print_to_screen()
    probs = []
    ptypes = []
    for i, (sett, files) in enumerate(zip(all_settings, all_files)):
        msgtxt = f'Analysing ensemble {i} of {nens}'
        # We change filetype to repptis instead of retis for anaylsis
        repptis_str = 'pathensemble_repptis'
        files0 = [(repptis_str, files[0][1])]
        print_to_screen(msgtxt, level='info')
        print_to_screen()
        if i == 0:
            result = run_analysis_files(sett, files0, plotter, txt_plotter)
            results['pathensemble0'] = result['pathensemble_repptis']
            report_txt = generate_report('retis0', result, output='txt')[0]
            print_to_screen(''.join(report_txt))
            print_to_screen()
        else:
            result = run_analysis_files(sett, files0, plotter, txt_plotter)
            results['pathensemble'].append(result['pathensemble_repptis'])
            pp_lmr = result[repptis_str]['out']['prun_sl'][-1]
            pp_rml = result[repptis_str]['out']['prun_sr'][-1]
            ptype = result[repptis_str]['out']['ptypes']
            ptypes.append(ptype)
            pp_lml = 1 - pp_lmr
            pp_rmr = 1 - pp_rml
            pp_lm_err = result[repptis_str]['out']['blockerror_sl'][2]
            pp_rm_err = result[repptis_str]['out']['blockerror_sr'][2]
            probs.append({"LMR": {"mean": pp_lmr,
                                  "std": pp_lm_err},
                          "RML": {"mean": pp_rml,
                                  "std": pp_rm_err},
                          "RMR": {"mean": pp_rmr,
                                  "std": pp_rm_err},
                          "LML": {"mean": pp_lml,
                                  "std": pp_lm_err}, })

            report_txt = generate_report('pptis', result, output='txt')[0]
            print_to_screen(''.join(report_txt))
            print_to_screen()

    # flux first:
    time_subcycles = settings['engine'].get('subcycles', 1)
    timestep = settings['engine']['timestep'] * time_subcycles
    flux, flux_error = retis_flux(results['pathensemble0']['out'],
                                  results['pathensemble'][0]['out'],
                                  timestep)
    results['flux'] = {'value': flux, 'error': flux_error,
                       'unit': units}

    results.update(repptis_running_pcross_analysis(ptypes))
    figures = plotter.output_matched_probability(
            [],
            settings['simulation']['interfaces'],
            results, reptis=True
    )
    txt_plotter.output_matched_probability(
            [],
            settings['simulation']['interfaces'],
            results, reptis=True
    )
    results['matched_fig'] = figures
    results['cross'] = {'value': results['pcrossrun'][-1],
                        'error': results['pcrossrun_hav_rele']}

    rate, rate_error = retis_rate(results['pcrossrun'][-1],
                                  results['pcrossrun_hav_rele'],
                                  flux, flux_error)
    results['rate'] = {'value': rate, 'error': rate_error,
                       'unit': units}

    xi, tau, perm, xi_err, tau_err, perm_err = perm_calculations(
        results['pathensemble0'], results['pcrossrun'][-1],
        results['pcrossrun_hav_rele'])
    xi_figs, tau_figs = perm_figures(results['pathensemble0'], plotter)
    results['xi'] = {'value': xi,
                     'error': xi_err,
                     'fig': xi_figs}
    results['tau'] = {'value': tau,
                      'error': tau_err,
                      'fig': tau_figs}
    results['perm'] = {'value': perm,
                       'error': perm_err}

    print_to_screen('Overall results', level='success')
    print_to_screen('===============', level='success')
    print_to_screen()
    print_value_error('REPPTIS Crossing probability',
                      results['pcrossrun'][-1],
                      results['pcrossrun_hav_rele'], level='success')
    print_to_screen()
    print_value_error(f'Initial flux (units 1/{units})', flux,
                      flux_error, level='success')
    print_to_screen()
    print_value_error(f'Rate constant (units 1/{units})', rate,
                      rate_error, level='success')
    if permeability:
        print_to_screen()
        print_value_error('Xi', xi, xi_err, level='success')
        print_to_screen()
        print_value_error(f'Tau/dz (unit: {units}/OP-unit)',
                          tau, tau_err, level='success')
        print_to_screen()
        print_value_error(f'Permeability (unit: OP-unit/{units})',
                          perm, perm_err, level='success')
    return results


def run_mdflux_analysis(settings, plotter, txt_plotter):
    """Run the analysis for the md flux simulation.

    Parameters
    ----------
    settings : dict
        The settings to use for an analysis/simulation.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    Returns
    -------
    out : list or dict or similar
        The output from the analysis.

    """
    sim = settings['simulation']
    sim_task = sim['task']
    exe_path = settings['simulation'].get('exe-path', os.getcwd())
    files = []
    for file_type in _FILES[sim_task]:
        filename = _FILES[sim_task][file_type]
        if os.path.isfile(os.path.join(exe_path, filename)):
            files.append((file_type, os.path.join(exe_path, filename)))
    msgtxt = 'Running analysis of a MD flux simulation...'
    print_to_screen(msgtxt, level='info')
    print_to_screen()
    result = run_analysis_files(settings, files, plotter, txt_plotter)
    report_txt = generate_report('md-flux', result, output='txt')[0]
    print_to_screen(''.join(report_txt))
    print_to_screen()
    return result


def run_analysis_files(settings, files, plotter, txt_plotter):
    """Run the analysis on a collection of files.

    Parameters
    ----------
    settings : dict
        This dict contains settings which dictate how the
        analysis should be performed and it should also contain
        information on how the simulation was performed.
    files : list of tuples
        This list contains the raw files to be analysed. The
        tuples are on format ('filetype', 'filename').
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.

    Returns
    -------
    out: dict
        The results from the analysis.

    """
    results = {}
    for (file_type, file_name) in files:
        result, raw_data = analyse_file(file_type, file_name, settings)
        figures = output_results(file_type, plotter, result, raw_data)
        txt_file = output_results(file_type, txt_plotter, result, raw_data)
        results[file_type] = {'out': result,
                              'figures': figures,
                              'txtfile': txt_file}
    return results


def read_first_block(file_type, file_name):
    """Read the first block of data from a file.

    This method will read the first block of data from a file and
    immediately return.

    Parameters
    ----------
    file_type : string
        A string used for selecting the file object to use here.
    file_name : string
        The file to open.

    Returns
    -------
    out : numpy.array
        The raw data read from the file.

    """
    first_block = None
    class_map = {'energy': EnergyFile, 'order': OrderFile, 'cross': CrossFile}
    if file_type not in class_map:
        raise ValueError('Unknown file type requested in analysis.')

    klass = class_map[file_type]

    with klass(file_name, 'r') as fileobj:
        for block in fileobj.load():
            if first_block is None:
                first_block = block
            else:
                logger.warning(
                    (
                        'Noticed a second block in the input file "%s".\n'
                        'This will be ignored by the flux analysis.\n'
                        'Are you are running the analysis with '
                        'the correct input?'
                    ),
                    file_name
                )
                break
    if first_block is None:
        return None
    return first_block['data']


def output_results(file_type, plotter, result, rawdata):
    """Plot the results.

    Plotting might here refer to both figures and text.

    Parameters
    ----------
    file_type : string
        This determines what we are going to plot.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    result : list of lists or dicts
        This contains the results from a specific analysis.
    rawdata : list of floats, lists or objects
        The raw data with analysis results.

    Returns
    -------
    out : list of strings
        These are the files that the plotter created.

    """
    if file_type == 'cross':
        return plotter.output_flux(result)
    if file_type == 'order':
        return plotter.output_orderp(result, rawdata)
    if file_type == 'energy':
        return plotter.output_energy(result, rawdata)
    if file_type == 'pathensemble':
        return plotter.output_path(result, rawdata)
    if file_type == 'pathensemble_repptis':
        return plotter.output_pppath(result, rawdata)


def analyse_file(file_type, file_name, settings):
    """Run analysis on the given file.

    This function is included for convenience so that we can call an
    analysis like ``analyse_file('cross', 'cross.txt')`` i.e. it should
    automatically open the file and apply the correct analysis
    according to a given file type. Here we return a function to do the
    analysis, so we are basically wrapping one of the analysis
    functions. This is done in case we wish to rerun the analysis but
    with different settings for instance.

    Parameters
    ----------
    file_type : string
        This is the type of file we are to analyse.
    file_name : string
        The file name to open.
    settings : dict
        This dict contains settings which dictate how the
        analysis should be performed and information on how the
        simulation was performed.

    Returns
    -------
    results : list or dict
        The output from the analysis
    raw_data : list, numpy.array or another type of object
        The raw data used in the analysis.

    """
    function = _ANALYSIS_FUNCTIONS.get(file_type, None)
    if function is None:
        msgtxt = f'Unknown file type "{file_type}" requested!'
        logger.error(msgtxt)
        raise ValueError(msgtxt)
    results, raw_data = None, None
    if file_type in ('energy', 'order', 'cross'):
        # Read the first block of raw data and analyse it:
        raw_data = read_first_block(file_type, file_name)
        results = function(raw_data, settings)
    elif file_type in ('pathensemble', 'pathensemble_repptis'):
        # Create a PathEnsembleFile object to use for the analysis.
        interfaces = settings['simulation']['interfaces']
        ensemble_settings = {
            'ensemble_number': settings['tis']['ensemble_number'],
            'interfaces': interfaces,
            'detect': settings['tis'].get('detect', interfaces[1])
        }
        raw_data = PathEnsembleFile(file_name, 'r', ensemble_settings)
        results = function(raw_data, settings)
    return results, raw_data


def analyse_and_output_matched(raw_data, plotter, txt_plotter,
                               settings, flux=None):
    """Analyse and output matched probability.

    This will calculate the over-all crossing probability by combining
    results from many path simulations.

    Parameters
    ----------
    raw_data : list, numpy.array or another type of object
        The raw data used in the analysis.
    plotter : object like :py:class:`.Plotter`
        This is the object that handles the plotting.
    txt_plotter : object like :py:class:`.Plotter`
        This is the object that handles the text output.
    settings : dict
        This dictionary contains settings for the analysis.
        Here we make use of the following keys from the
        analysis section:

        * `ngrid`: The number of grid points for calculating the
          crossing probability as a function of the order parameter.
        * `maxblock`: The max length of the blocks for the block error
          analysis. Note that this will maximum be equal the half of the
          length of the data, see `block_error` in `.analysis`.
        * `blockskip`: Can be used to skip certain block lengths.
          A `blockskip` equal to `n` will consider every n'th block up
          to `maxblock`, i.e. it will use block lengths equal to `1`,
          `1+n`, `1+2n`, etc.
        * `bins`: The number of bins to use for creating histograms.

    flux : float, optional
        The computed flux to be used only in RETIS if the running average
        of the computed rate is desired.

    Returns
    -------
    out[0] : dict
        This dict contains the results from the analysis
    out[1] : list of dicts
        A dictionary with the figure files created (if any).
    out[2] : list of strings
        A list with the text files created (if any).

    """
    path_results, ensemble_names, detect = [], [], []
    interface_left = None
    for ensemble in raw_data:
        path_results.append(ensemble['out'])
        ensemble_names.append(ensemble['out']['ensemble'])
        detect.append(ensemble['out']['detect'])
        if interface_left is None:
            interface_left = ensemble['out']['interfaces'][0]
    result = match_probabilities(path_results, detect, settings)
    if flux is not None:
        result['overall-rrun'] = flux*result['overall-prun']
    # for the figure, we add the A interface:
    detect_plot = [interface_left] + detect
    figures = plotter.output_matched_probability(ensemble_names,
                                                 detect_plot,
                                                 result)
    outtxt = txt_plotter.output_matched_probability(ensemble_names,
                                                    detect,
                                                    result)
    return result, figures, outtxt


def perm_figures(results0, plotter):
    out = results0['out']
    # Quick check to see if we need to do something
    if 'xi' not in out:
        return None, None
    xi = out['xi']
    xierror = out['xierror']

    bin_edges = out['tau_bin']
    tau = out['tau']
    tauerror = out['tauerror']
    tau_ref = out['tau_ref']
    tau_ref_bins = out['tau_ref_bins']
    xi_figures = plotter.output_xi(xi, xierror)
    tau_figures = plotter.output_tau(tau, tauerror, bin_edges,
                                     tau_ref, tau_ref_bins)
    return xi_figures, tau_figures


def repptis_running_pcross_analysis(l_ptypes):
    """Recursive block analysis for repptis.

    Parameters
    ----------
    l_ptypes : list
        List of cumsum ptypes for the ensembles [0+-'], [1+-], ... [N-1+-].

    Returns
    -------
    results : dict
        The results from the analysis.

    """
    l_locps = []
    for ptypes in l_ptypes:  # loop over ensembles
        # if a division by zero occurs, we set the probability to NaN
        # casting to floats for true_divide
        ptypes = ptypes.astype(float)
        pmps = np.divide(ptypes[:, 0], ptypes[:, 0]+ptypes[:, 1],
                         out=np.full_like(ptypes[:, 0], np.nan),
                         where=(ptypes[:, 0]+ptypes[:, 1]) != 0)
        pmms = np.divide(ptypes[:, 1], ptypes[:, 0]+ptypes[:, 1],
                         out=np.full_like(ptypes[:, 0], np.nan),
                         where=(ptypes[:, 0]+ptypes[:, 1]) != 0)
        ppps = np.divide(ptypes[:, 2], ptypes[:, 2]+ptypes[:, 3],
                         out=np.full_like(ptypes[:, 2], np.nan),
                         where=(ptypes[:, 2]+ptypes[:, 3]) != 0)
        ppms = np.divide(ptypes[:, 3], ptypes[:, 2]+ptypes[:, 3],
                         out=np.full_like(ptypes[:, 2], np.nan),
                         where=(ptypes[:, 2]+ptypes[:, 3]) != 0)
        # we have to_list them, as we will append them until all have N_max cyc
        l_locps.append([pmms.tolist(), pmps.tolist(),
                        ppps.tolist(), ppms.tolist()])
    # So, l_locps: list of lists of lists: n_ensembles x 4 x n_blocks
    # extend the l_locps to the same length. We copy the last element of the
    # list until the length is the same as the longest list
    max_len = max([len(ll[0]) for ll in l_locps])
    for ll in l_locps:
        while len(ll[0]) < max_len:
            for i in range(4):
                ll[i].append(ll[i][-1])
    # We can now calculate the global crossing probabilities, for each cycle
    pcross_run = []
    for icyc in range(max_len):
        # We have to get the local crossing probabilities for each ensemble
        l_pcross = [[ll[i][icyc] for i in range(4)] for ll in l_locps]
        # so l_pcross is shape n_ensembles x 4, but we need 4 lists of
        # length n_ensembles
        l_pcross = list(map(list, zip(*l_pcross)))
        # Now we calculate the global crossing probabilities
        _, _, pcross = get_global_probz(*l_pcross)
        # We append the last element of pcross to the running list
        pcross_run.append(pcross[-1] if pcross is not np.nan else np.nan)
    # Now do the recursive block analysis
    rel_errors, nstatineff, half_av_err = recursive_block_analysis(pcross_run)

    results = {'pcross': pcross,
               'pcrossrun': pcross_run,
               'pcrossrun_rele': rel_errors,
               'nstatineff': nstatineff,
               'pcrossrun_hav_rele': half_av_err}

    # make into retis format
    results['overall-cycle'] = list(range(len(pcross_run)))
    results['overall-prun'] = pcross_run
    results['overall-error'] = [list(range(len(rel_errors))), 0, 0,
                                rel_errors, rel_errors[-1]]
    return results


def get_global_probz(pmps, pmms, ppps, ppms):
    """Return the global crossing probabilities for the PPTIS simulation.

    This follows the recursive relation
    P_plus[j] = (pLMR[j-1]*P_plus[j-1]) / (pLMR[j-1]+pLML[j-1]*P_min[j-1])
    P_min[j] = (pRML[j-1]P_min[j-1])/(pLMR[j-1]+pLML[j-1]*P_min[j-1])
    where P_plus[0] = 1, P_min[0] = 1, and j is the ensemble number.

    Parameters
    ----------
    pmps : list of floats
        The local probability (P) of having type LMR (MinusPlus) for each
        ensemble (S).
    pmms : list of floats
        Local LML
    ppps : list of floats
        Local RMR
    ppms : list of floats
        Local RML

    Returns
    -------
    pmin : list of floats
        Float per ensemble i. Represents the probability of crossing A earlier
        than i+1, given that the path crossed i.
    pplus : list of floats
        Float per ensemble i. Represents the probability of crossing i+1
        earlier than A, given that the path crossed i.
    pcross : list of floats
        Float per ensemble i. Represents the TIS probability of crossing i+1.
    """
    # if there is any NaN in pmps, pmms, ppps, ppms, return NaN
    if np.isnan(pmps).any() or np.isnan(pmms).any() or \
            np.isnan(ppps).any() or np.isnan(ppms).any():
        return [np.nan, np.nan, np.nan]
    pplus, pmin, pcross = [1.], [1.], [1., pmps[0]]
    for i, pmp, pmm, _, ppm in zip(range(len(pmps)), pmps, pmms, ppps, ppms):
        if i == 0:  # This is [0^{\pm}'], so skip
            continue
        # Calculate the global crossing probabilities
        # if divide by zero, or divide by NaN, return NaN
        if pmp + pmm * pmin[-1] == 0 or\
                (pmp is np.nan) or\
                (pmm * pmin[-1] is np.nan):
            return [np.nan, np.nan, np.nan]
        pplus.append((pmp*pplus[-1])/(pmp+pmm*pmin[-1]))
        pmin.append((ppm*pmin[-1])/(pmp+pmm*pmin[-1]))
        # Calculate the TIS probabilities
        pcross.append(pmps[0]*pplus[-1])
    return pmin, pplus, pcross


def recursive_block_analysis(flist, minblocks=5):
    """Run a recursive block analysis on (time-series) data f.

    Parameters
    ----------
    flist : list of floats
        The data to analyse.
    minblocks : int, optional
        The minimum number of blocks to use for the analysis.

    Returns
    -------
    rel_errors : list of floats
        The relative errors for each block length.
    nstatineff : float
        The statistical inefficiency.
    half_av_err : float
        The average error for the second half of the relative errors.

    """
    # flist is a list of floats, but we need to remove the np.nan's and 0's
    flist = np.array(flist)
    flist = flist[~np.isnan(flist)]
    flist = flist[flist != 0]
    # We now run the recursive blocking analysis
    # If we don't have enough data, we don't want the entire analysis to
    # fail. So just return NaNs, such that we still get those beautiful local
    # plots in the report PDF.
    if len(flist) < minblocks:
        return [np.nan], np.nan, np.nan
    bestavg = flist[-1]
    maxblocklen = int(len(flist)/minblocks)
    rel_errors = []
    for blocklen in range(1, maxblocklen+1):
        runav_red = flist[blocklen-1::blocklen]
        blocks = recursive_blocks(runav_red)
        sum_quad_diff = np.sum(np.fromiter(((b-bestavg)**2 for b in blocks),
                               dtype=float))
        nb = len(blocks)
        aerr2 = sum_quad_diff/(nb*(nb-1))
        aerr = np.sqrt(aerr2)
        rerr = aerr/bestavg
        rel_errors.append(rerr)

    # We will now take the mean of the second half of the relative errors
    half_av_err = np.mean(rel_errors[int(len(rel_errors)/2):])
    nstatineff = (half_av_err/rel_errors[0])**2
    # Now we have to add the NaN's back in, as 0's
    return rel_errors, nstatineff, half_av_err


def recursive_blocks(recu):
    blocks = []
    for i in range(len(recu)):
        if i == 0:
            blocks.append(recu[i])
        else:
            blocks.append((i + 1) * recu[i] - i * recu[i - 1])
    return blocks



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of some common methods that might be useful.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

bit_fat_comparer (:py:func`.big_fat_comparer`)
    Method to compare two nested list/dictionaries.

compare_objects (:py:func`.compare_objects`)
    Method to compare two PyRETIS objects.

counter (:py:func`.counter`)
    Function to count the number of iterations.

crossing_counter (:py:func`.crossing_counter`)
    Function to count the crossing of a path on an interface.

crossing_finder (:py:func`.crossing_finder`)
    Function to get the shooting points of the crossing of a path
    on an interface.

import_from (:py:func:`.import_from`)
    A method to dynamically import method/classes etc. from user
    specified modules.

inspect_function (:py:func:`.inspect_function`)
    A method to obtain information about arguments, keyword arguments
    for functions.

initiate_instance (:py:func:`.initiate_instance`)
    Method to initiate a class with optional arguments.

generic_factory (:py:func:`.generic_factory`)
    Create instances of classes based on settings.

null_move (:py:func`.compare_objects`)
    Method to do not move.

priority_checker (:py:func`.priority_checker`)
    Method to check ensemble to prioritize.

relative_shoots_select (:py:func`.compare_objects`)
    Method to select the shooting ensemble.

segments_counter (:py:func`.segments_counter`)
    Function that counts the number of segments between two interfaces.

trim_path_between_interfaces (:py:func`.trim_path_between_interfaces`)
    Function to trim a path between interfaces.

select_and_trim_a_segment (:py:func`.select_and_trim_a_segment`)
    Function to trim a path between interfaces plus the two external points.

compute_weight (:py:func:`.compute_weight`)
    A method to compute the statistical weight of a path generated by a
    Stone Skipping and Wire Fencing move.

soft_partial_exit (:py:func`.soft_partial_exit`)
    Function that check the presence of the EXIT file,
    and kindly stops the iterator.

"""
import logging
import inspect
import importlib
import os
import sys
import numpy as np
from pyretis.inout import print_to_screen

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = ['import_from', 'inspect_function', 'initiate_instance',
           'generic_factory', 'crossing_counter', 'crossing_finder',
           'segments_counter', 'select_and_trim_a_segment', 'counter',
           'trim_path_between_interfaces', 'big_fat_comparer',
           'soft_partial_exit', 'null_move', 'relative_shoots_select',
           'compute_weight', 'priority_checker', 'compare_objects']


def counter():
    """Return how many times this function is called."""
    counter.count = 0 if not hasattr(counter, 'count') else counter.count + 1
    return counter.count


def import_from(module_path, function_name):
    """Import a method/class from a module.

    This method will dynamically import a specified method/object
    from a module and return it. If the module can not be imported or
    if we can't find the method/class in the module we will raise
    exceptions.

    Parameters
    ----------
    module_path : string
        The path/filename to load from.
    function_name : string
        The name of the method/class to load.

    Returns
    -------
    out : object
        The thing we managed to import.

    """
    try:
        module_name = os.path.basename(module_path)
        module_name = os.path.splitext(module_name)[0]
        spec = importlib.util.spec_from_file_location(module_name,
                                                      module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module_name] = module
        logger.debug('Imported module: %s', module)
        return getattr(module, function_name)
    except (ImportError, IOError):
        msg = f'Could not import module: {module_path}'
        logger.critical(msg)
    except AttributeError:
        msg = f'Could not import "{function_name}" from "{module_path}"'
        logger.critical(msg)
    raise ValueError(msg)


def _arg_kind(arg):
    """Determine kind for a given argument.

    This method will help :py:func:`.inspect_function` to determine
    the correct kind for arguments.

    Parameters
    ----------
    arg : object like :py:class:`inspect.Parameter`
        The argument we will determine the type of.

    Returns
    -------
    out : string
        A string we use for determine the kind.

    """
    kind = None
    if arg.kind == arg.POSITIONAL_OR_KEYWORD:
        if arg.default is arg.empty:
            kind = 'args'
        else:
            kind = 'kwargs'
    elif arg.kind == arg.POSITIONAL_ONLY:
        kind = 'args'
    elif arg.kind == arg.VAR_POSITIONAL:
        kind = 'varargs'
    elif arg.kind == arg.VAR_KEYWORD:
        kind = 'keywords'
    elif arg.kind == arg.KEYWORD_ONLY:
        # We treat these as keyword arguments:
        kind = 'kwargs'
    return kind


def big_fat_comparer(any1, any2, hard=False):
    """Check if two dictionary are the same, regardless their complexity.

    Parameters
    ----------
    any1 : anything
    any2 : anything
    hard : boolean, optional
        Raise ValueError if any1 and any2 are different

    Returns
    -------
    out : boolean
        True if any1 = any2, false otherwise

    """
    if type(any1) is not type(any2):
        if hard:
            raise ValueError('Fail type', any1, any2)
        return False

    if isinstance(any1, (list, tuple)):
        if len(any1) != len(any2):
            if hard:
                raise ValueError('Fail list length', any1, any2)
            return False

        for key1, key2 in zip(any1, any2):
            if not big_fat_comparer(key1, key2, hard):
                if hard:
                    raise ValueError('Fail item in list',
                                     any1, any2)  # pragma: no cover
                return False

    elif isinstance(any1, np.ndarray):
        if any1.shape != any2.shape:
            if hard:
                raise ValueError('Fail np array shape', any1, any2)
            return False

        for key1, key2 in zip(np.nditer(any1), np.nditer(any2)):
            if not (key1 == key2).all():
                if hard:
                    raise ValueError('Fail np array item', any1, any2)
                return False

    elif isinstance(any1, dict):
        for key in any1:
            if key not in any2:
                if hard:
                    raise ValueError('Fail dict', any1, any2)
                return False

            if not isinstance(any1[key], type(any2[key])):
                if hard:
                    raise ValueError('Fail types', any1[key], any2[key])
                return False

            if isinstance(any1[key], (dict, list, tuple, np.ndarray)):
                if not big_fat_comparer(any1[key], any2[key], hard):
                    if hard:
                        raise ValueError('Fail item',
                                         any1[key],
                                         any2[key])  # pragma: no cover
                    return False

            else:
                if any1[key] != any2[key]:
                    if hard:
                        raise ValueError('Fail item', any1[key], any2[key])
                    return False

        for key in any2:
            if key not in any1:
                if hard:
                    raise ValueError('Fail item', any1, any2)
                return False

    else:
        if any1 != any2:
            if hard:
                raise ValueError('Fail item', any1, any2)
            return False

    return True


def inspect_function(function):
    """Return arguments/kwargs of a given function.

    This method is intended for use where we are checking that we can
    call a certain function. This method will return arguments and
    keyword arguments a function expects. This method may be fragile -
    we assume here that we are not really interested in args and
    kwargs and we do not look for more information about these here.

    Parameters
    ----------
    function : callable
        The function to inspect.

    Returns
    -------
    out : dict
        A dict with the arguments, the following keys are defined:

        * `args` : list of the positional arguments
        * `kwargs` : list of keyword arguments
        * `varargs` : list of arguments
        * `keywords` : list of keyword arguments

    """
    out = {'args': [], 'kwargs': [],
           'varargs': [], 'keywords': []}
    arguments = inspect.signature(function)  # pylint: disable=no-member
    for arg in arguments.parameters.values():
        kind = _arg_kind(arg)
        if kind is not None:
            out[kind].append(arg.name)
        else:  # pragma: no cover
            logger.critical('Unknown variable kind "%s" for "%s"',
                            arg.kind, arg.name)
    return out


def _pick_out_arg_kwargs(klass, settings):
    """Pick out arguments for a class from settings.

    Parameters
    ----------
    klass : class
        The class to initiate.
    settings : dict
        Positional and keyword arguments to pass to `klass.__init__()`.

    Returns
    -------
    out[0] : list
        A list of the positional arguments.
    out[1] : dict
        The keyword arguments.

    """
    info = inspect_function(klass.__init__)
    used, args, kwargs = set(), [], {}
    for arg in info['args']:
        if arg == 'self':
            continue
        try:
            args.append(settings[arg])
            used.add(arg)
        except KeyError:
            msg = f'Required argument "{arg}" for "{klass}" not found!'
            raise ValueError(msg)
    for arg in info['kwargs']:
        if arg == 'self':
            continue
        if arg in settings:
            kwargs[arg] = settings[arg]
    return args, kwargs


def initiate_instance(klass, settings):
    """Initialise a class with optional arguments.

    Parameters
    ----------
    klass : class
        The class to initiate.
    settings : dict
        Positional and keyword arguments to pass to `klass.__init__()`.

    Returns
    -------
    out : instance of `klass`
        Here, we just return the initiated instance of the given class.

    """
    args, kwargs = _pick_out_arg_kwargs(klass, settings)
    # Ready to initiate:
    msg = 'Initiated "%s" from "%s" %s'
    name = klass.__name__
    mod = klass.__module__
    if not args:
        if not kwargs:
            logger.debug(msg, name, mod, 'without arguments.')
            return klass()
        logger.debug(msg, name, mod, 'with keyword arguments.')
        return klass(**kwargs)
    if not kwargs:
        logger.debug(msg, name, mod, 'with positional arguments.')
        return klass(*args)
    logger.debug(msg, name, mod,
                 'with positional and keyword arguments.')
    return klass(*args, **kwargs)


def generic_factory(settings, object_map, name='generic'):
    """Create instances of classes based on settings.

    This method is intended as a semi-generic factory for creating
    instances of different objects based on simulation input settings.
    The input settings define what classes should be created and
    the object_map defines a mapping between settings and the
    class.

    Parameters
    ----------
    settings : dict
        This defines how we set up and select the order parameter.
    object_map : dict
        Definitions on how to initiate the different classes.
    name : string, optional
        Short name for the object type. Only used for error messages.

    Returns
    -------
    out : instance of a class
        The created object, in case we were successful. Otherwise, we
        return none.

    """
    try:
        klass = settings['class'].lower()
    except KeyError:
        msg = 'No class given for %s -- could not create object!'
        logger.critical(msg, name)
        return None
    if klass not in object_map:
        logger.critical('Could not create unknown class "%s" for %s',
                        settings['class'], name)
        return None
    cls = object_map[klass]['cls']
    return initiate_instance(cls, settings)


def numpy_allclose(val1, val2):
    """Compare two values with all close from numpy.

    Here, we allow for one, or both, of the values to be None.
    Note that if val1 == val2 but are not of a type known to
    numpy, the returned value will be False.

    Parameters
    ----------
    val1 : np.array
        The variable in the comparison.
    val2 : np.array
        The second variable in the comparison.

    Returns
    -------
    out : boolean
        True if the values are equal, False otherwise.

    """
    if val1 is None and val2 is None:
        return True
    if val1 is None and val2 is not None:
        return False
    if val1 is not None and val2 is None:
        return False
    try:
        return np.allclose(val1, val2)
    except TypeError:
        return False


def compare_objects(obj1, obj2, attrs, numpy_attrs=None):
    """Compare two PyRETIS objects.

    This method will compare two PyRETIS objects by checking
    the equality of the attributes. Some of these attributes
    might be numpy arrays in which case we use the
    :py:function:`.numpy_allclose` defined in this module.

    Parameters
    ----------
    obj1 : object
        The first object for the comparison.
    obj2 : object
        The second object for the comparison.
    attrs : iterable of strings
        The attributes to check.
    numpy_attrs : iterable of strings, optional
        The subset of attributes which are numpy arrays.

    Returns
    -------
    out : boolean
        True if the objects are equal, False otherwise.

    """
    if not obj1.__class__ == obj2.__class__:
        logger.debug(
            'The classes are different %s != %s',
            obj1.__class__, obj2.__class__
        )
        return False
    if not len(obj1.__dict__) == len(obj2.__dict__):
        logger.debug('Number of attributes differ.')
        return False
    # Compare the requested attributes:
    for key in attrs:
        try:
            val1 = getattr(obj1, key)
            val2 = getattr(obj2, key)
        except AttributeError:
            logger.debug('Failed to compare attribute "%s"', key)
            return False
        if numpy_attrs and key in numpy_attrs:
            if not numpy_allclose(val1, val2):
                logger.debug('Attribute "%s" differ.', key)
                return False
        else:
            if not val1 == val2:
                logger.debug('Attribute "%s" differ.', key)
                return False
    return True


def null_move(ensemble, cycle):
    """Perform a null move for a path ensemble.

    The null move simply consist of accepting the last accepted path
    again.

    Parameters
    ----------
    ensemble: dict, it contains:
      *  path_ensemble : object like :py:class:`.PathEnsemble`
             This is the path ensemble to update with the null move
    cycle : integer
        The current cycle number

    Returns
    -------
    out[0] : boolean
        Should the path be accepted or not? Here, it's always True
        since the null move is always accepted.
    out[1] : object like :py:class:`.PathBase`
        The unchanged path.
    out[2] : string
        The status will here be 'ACC' since we just accept
        the last accepted path again in this move.

    """
    path_ensemble = ensemble['path_ensemble']
    logger.info('Null move for: %s', path_ensemble.ensemble_name)
    status = 'ACC'
    path = path_ensemble.last_path
    if not path.get_move() == 'ld':
        path.set_move('00')
    path_ensemble.add_path_data(path, status, cycle=cycle)
    return True, path, status


def relative_shoots_select(ensembles, rgen, relative):
    """Randomly select the ensemble for 'relative' shooting moves.

    Here we select the ensemble to do the shooting in based on relative
    probabilities. We draw a random number between 0 and 1 which is used to
    select the ensemble.

    Parameters
    ----------
    ensembles : list of objects like :py:class:`.PathEnsemble`
        This is a list of the ensembles we are using in the RETIS method.
    rgen : object like :py:class:`.RandomGenerator`
        This is a random generator. Here we assume that we can call
        `rgen.rand()` to draw random uniform numbers.
    relative : list of floats
        These are the relative probabilities for the ensembles. We
        assume here that these numbers are normalised.

    Returns
    -------
    out[0] : integer
        The index of the path ensemble to shoot in.
    out[1] : object like :py:class:`.PathEnsemble`
        The selected path ensemble for shooting.

    """
    freq = rgen.rand()
    cumulative = 0.0
    idx = None
    for i, path_freq in enumerate(relative):
        cumulative += path_freq
        if freq < cumulative:
            idx = i
            break
    try:
        ensemble = ensembles[idx]
    except TypeError:
        raise ValueError('Error in relative shoot frequencies! Aborting!')
    return idx, ensemble


def segments_counter(path, interface_l, interface_r, reverse=False):
    """Count the directional segment between interfaces.

    Method to count the number of the directional segments of the path,
    along the orderp, that connect FROM interface_l TO interface_r.

    Parameters
    -----------
    path : object like :py:class:`.PathBase`
        This is the input path which segments will be counted.
    interface_r : float
        This is the position of the RIGHT interface.
    interface_l : float
        This is the position of the LEFT interface.
    reverse : boolean, optional
        Check on a reversed path.

    Returns
    -------
    n_segments : integer
        Segment counter

    """
    icros, n_segments = -1, 0
    for i in range(path.length - 1):
        op1 = path.phasepoints[i].order[0]
        op2 = path.phasepoints[i+1].order[0]
        if reverse and op1 >= interface_r > op2 or\
                not reverse and op2 > interface_l >= op1:
            icros = i
        if reverse and op1 >= interface_l > op2 or\
                not reverse and op2 > interface_r >= op1:
            if icros != -1:
                icros = -1
                n_segments += 1
    return n_segments


def crossing_counter(path, interface):
    """Count the crossing to an interfaces.

    Method to count the crosses of a path over an interface.

    Parameters
    -----------
    path : object like :py:class:`.PathBase`
        Input path which will be trimmed.
    interface : float
        The position of the interface.

    Returns
    -------
    cnt : integer
        Number of crossing of the given interface.

    """
    cnt = 0
    for i in range(len(path.phasepoints[:-1])):
        op1 = path.phasepoints[i].order[0]
        op2 = path.phasepoints[i+1].order[0]
        if op2 >= interface > op1 or op1 >= interface > op2:
            cnt += 1
    return cnt


def crossing_finder(path, interface, last_frame=False):
    """Find the crossing to an interfaces.

    Method to select the crosses of a path over an interface.

    Parameters
    -----------
    path : object like :py:class:`.PathBase`
        Input path which will be trimmed.
    interface : float
        Interface position.
    last_frame : boolean, optional
        Determines if the last crossing will be selected or not.

    Returns
    -------
    ph1, ph2 : snapshots
        Define the randomly picked crossing,
        one right before and one right after the interface.

    """
    ph1, ph2 = [], []
    for i in range(len(path.phasepoints[:-1])):
        op1 = path.phasepoints[i].order[0]
        op2 = path.phasepoints[i+1].order[0]
        if op2 >= interface > op1 or op1 >= interface > op2:
            ph1.append(path.phasepoints[i])
            ph2.append(path.phasepoints[i+1])
    if not ph1:
        return None, None
    assert ph1, 'No crossing point available'
    idx = -1 if last_frame else path.rgen.random_integers(0, len(ph1) - 1)
    return ph1[idx], ph2[idx]


def trim_path_between_interfaces(path, interface_l, interface_r):
    """Cut a path between the two interfaces.

    The method cut a path and keeps only what is within the range
    (interface_l interface_r).
    -Be careful, it can provide multiple discontinuous segments-
    =Be carefull2 consider if you need to make this check left inclusive
    (as the ensemble should be left inclusive)

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be trimmed.
    interface_r : float
        This is the position of the RIGHT interface.
    interface_l : float
        This is the position of the LEFT interface.

    Returns
    -------
    new_path : object like :py:class:`.PathBase`
        This is the output trimmed path.

    """
    new_path = path.empty_path()
    for phasepoint in path.phasepoints:
        orderp = phasepoint.order[0]
        if interface_r > orderp > interface_l:
            new_path.append(phasepoint)
    new_path.maxlen = path.maxlen
    new_path.status = path.status
    new_path.time_origin = path.time_origin
    new_path.generated = 'ct'
    new_path.rgen = path.rgen
    return new_path


def select_and_trim_a_segment(path, interface_l, interface_r,
                              segment_to_pick=None):
    """Cut a directional segment from interface_l to interface_r.

    It keeps what is within the range [interface_l interface_r)
    AND the snapshots just after/before the interface.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be trimmed.
    interface_r : float
        This is the position of the RIGHT interface.
    interface_l : float
        This is the position of the LEFT interface.
    segment_to_pick : integer (n.b. it starts from 0)
        This is the segment to be selected, None = random

    Returns
    -------
    segment : object like :py:class:`.PathBase`
        a path segment composed only the snapshots for which
        orderp is between interface_r and interface_l and the
        ones right after/before the interfaces.

    """
    key = False
    segment = path.empty_path()
    segment_i = -1
    isave = -1
    if segment_to_pick is None:
        segment_number = segments_counter(path, interface_l, interface_r)
        segment_to_pick = path.rgen.random_integers(0, segment_number)

    for i, phasepoint in enumerate(path.phasepoints[:-1]):
        op1 = path.phasepoints[i].order[0]
        op2 = path.phasepoints[i+1].order[0]
        # NB: these are directional crossing
        if op2 >= interface_l > op1:
            # We are in the good region, segment_i
            if not key:
                segment_i += 1
            key = True
        if key:
            if segment_i == segment_to_pick:
                segment.append(phasepoint)
                isave = i
        if op2 >= interface_r > op1:
            if key and segment_i == segment_to_pick:
                segment.append(path.phasepoints[i+1])
            key = False

    if segment.length == 1:
        segment.append(path.phasepoints[isave + 1])
    segment.maxlen = path.maxlen
    segment.status = path.status
    segment.time_origin = path.time_origin
    segment.generated = 'sg'
    segment.rgen = path.rgen
    return segment


def wirefence_weight_and_pick(path, intf_l, intf_r, return_seg=False):
    """Calculate the weight of a path generated by the Wire Fence move.

    The WF path weight is determined by the total sum of valid sub-path
    phasepoints, where valid WF subpaths are defined as intf_l-intf_l,
    intf_l-intf_r and intf_r-intf_l sub-paths.

    if return_seg = True, a random valid WF sub-path is also returned.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be trimmed.
    intf_r : float
        This is the position of the RIGHT interface.
    intf_l : float
        This is the position of the LEFT interface.
    return_seg : boolean, optional
        Determines if a random valid WF sub-path is returned or not.

    Returns
    -------
    n_frames: int
        The weight of the path.
    segment : object like :py:class:`.PathBase`
        False (if return_seg=False) else a random valid WF sub-path.

    """
    key_l, key_r = False, False
    path_arr = []
    segment = False

    for i in range(len(path.phasepoints[:-1])):
        op1 = path.phasepoints[i].order[0]
        op2 = path.phasepoints[i+1].order[0]

        if (op1 < intf_l and op2 >= intf_r) or \
           (op2 < intf_l and op1 >= intf_r):
            pass
        elif op2 >= intf_l > op1 and not key_l:
            isave, key_l = i, True
        elif op2 < intf_r <= op1 and not key_r:
            isave, key_r = i, True
        elif key_r and op2 >= intf_r > op1:
            key_l, key_r = False, False
        elif True in (key_l, key_r) and (op2 < intf_l <= op1 or
                                         op2 >= intf_r > op1):
            key_l, key_r = False, False
            path_arr.append((isave, i+1, i-isave))

    n_frames = sum([i[2] for i in path_arr]) if path_arr else 0
    if return_seg and n_frames:
        sum_frames = 0
        subpath_select = path.rgen.rand()
        for i in path_arr:
            sum_frames += i[2]
            if sum_frames/n_frames >= subpath_select:
                segment = path.empty_path()
                for j in range(i[0], i[1]+1):
                    segment.append(path.phasepoints[j])
                segment.maxlen = path.maxlen
                segment.status = path.status
                segment.time_origin = path.time_origin
                segment.generated = 'ct'
                segment.rgen = path.rgen
                break

    return n_frames, segment


def compute_weight(path, interfaces, move):
    """Compute the High Acceptance path weight after a MC move.

    This function computes the weights that will be used in the
    computation of the P cross. This trick allows the use of
    the High Acceptance version of Stone Skipping or Wire Fencing,
    allowing the acceptance of B to A paths.
    The drawback is that swapping moves needs
    to account also for this different weights.
    The weight 1 will be returned for a path not generated by SS or WF.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        This is the input path which will be checked.
    interfaces : list/tuple of floats
        These are the interface positions of the form
        ``[left, middle, right]``.
    move : string, optional
        The MC move to compute the weights for.

    Returns
    -------
    out[0] : float
        The weight of the path.

    """
    weight = 1.

    if move == 'ss':
        weight = 1.*crossing_counter(path, interfaces[1])
    elif move == 'wf':
        wf_weight, _ = wirefence_weight_and_pick(path,
                                                 interfaces[1],
                                                 interfaces[2])
        weight = 1.*wf_weight

    if path.get_start_point(interfaces[0], interfaces[2]) != \
            path.get_end_point(interfaces[0], interfaces[2]):
        if move in ('ss', 'wf'):
            weight *= 2

    return weight


def priority_checker(ensembles, settings):
    """Determine the shooting ensemble during a RETIS simulation.

    Here we check whether to do priority shooting or not. If True,
    we either shoot from the ensemble with the fewest paths or
    ensemble [0^-] if all ensembles have the same no. of paths.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        Lit of dict of ensembles we are using in a path method.
    settings : dict
        This dict contains the settings for the RETIS method.

    Returns
    -------
    out[0] : list
        Returns a list of boolean dictating whether certain
        ensembles are to be skipped or not.

    """
    priority = settings.get('simulation', {}).get('priority_shooting', False)
    prio_skip = [False] * len(ensembles)
    if priority:
        lst_cycles = [ens['path_ensemble'].nstats['npath']
                      for ens in ensembles]
        # Are all ensemble npath values the same?
        if any(i != lst_cycles[0] for i in lst_cycles):
            # If not, let's make a list:
            prio_skip = [i == max(lst_cycles) for i in lst_cycles]
    return prio_skip


def soft_partial_exit(exe_path=''):
    """Check the presence of the EXIT file.

    Parameters
    ----------
    exe_path: string, optional
        Path for the EXIT file.

    Returns
    -------
    out : boolean
        True if EXIT is present.
        False if EXIT in not present.

    """
    exit_file = 'EXIT'
    if exe_path:
        exit_file = os.path.join(exe_path, exit_file)
    if os.path.isfile(exit_file):
        logger.info('Exit file found - will exit between steps.')
        print_to_screen('Exit file found - will exit between steps.',
                        level='warning')

        return True
    return False



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Compare energy output for CP2K."""
import sys
import colorama
from matplotlib import pyplot as plt
import numpy as np
from pyretis.inout import print_to_screen


plt.style.use('seaborn-v0_8-poster')


def main(cp2k_file1, cp2k_file2, plot=False):
    """Perform the test."""
    print_to_screen(f'Reading energy file: {cp2k_file1}',
                    level='info')
    energy_cp2k1 = np.loadtxt(cp2k_file1)
    print_to_screen(f'Reading energy file: {cp2k_file2}',
                    level='info')
    energy_cp2k2 = np.loadtxt(cp2k_file2)

    mse_ok = obtain_mses(energy_cp2k1, energy_cp2k2)

    if plot:
        print_to_screen('\nPlotting for comparison', level='message')
        plot_comparison(energy_cp2k1, energy_cp2k2, cp2k_file1, cp2k_file2)

    if not mse_ok:
        print_to_screen('\nComparison failed!', level='error')
        sys.exit(1)


def obtain_mses(energy_cp2k1, energy_cp2k2, tol=1.0e-5):
    """Obtain some mean squared errors."""
    if not energy_cp2k1.shape == energy_cp2k2.shape:
        return False
    _, ncol = energy_cp2k1.shape
    for key in range(1, ncol - 1):  # last column is used time, skip it!
        term1 = (energy_cp2k1[key] - energy_cp2k2[key])**2
        term2 = (np.average(energy_cp2k2[key]) - energy_cp2k2[key])**2
        rse = term1.sum() / term2.sum()
        level = 'info'
        tol_ok = True
        if tol:
            tol_ok = abs(rse) < tol
            if not tol_ok:
                level = 'error'
        print_to_screen(f'RES {key}: {rse}', level=level)
        if not tol_ok:
            return False
    return True


def plot_comparison(energy_cp2k1, energy_cp2k2, cp2k_file1, cp2k_file2):
    """Just plot some properties for the paths."""
    fig1 = plt.figure()
    ax11 = fig1.add_subplot(221)
    ax12 = fig1.add_subplot(222)
    ax21 = fig1.add_subplot(223)
    ax22 = fig1.add_subplot(224)
    plabel = f'CP2K ({cp2k_file1})'
    glabel = f'cp2k ({cp2k_file2})'

    ax11.plot(energy_cp2k1[:, 1], energy_cp2k1[:, 4],
              lw=4, ls='-', marker='o', label=plabel)
    ax11.plot(energy_cp2k2[:, 1], energy_cp2k2[:, 4],
              lw=2, ls='--', marker='^', label=glabel)
    ax11.legend()
    ax11.set_ylabel('Potential')

    ax12.plot(energy_cp2k2[:, 1], energy_cp2k2[:, 2],
              lw=4, ls='-', marker='o')
    ax12.plot(energy_cp2k2[:, 1], energy_cp2k2[:, 2],
              lw=2, ls='--', marker='^')
    ax12.set_ylabel('Kinetic')

    ax21.plot(energy_cp2k1[:, 1], energy_cp2k1[:, 5],
              lw=4, ls='-', marker='o')
    ax21.plot(energy_cp2k2[:, 1], energy_cp2k2[:, 5],
              lw=2, ls='--', marker='^')
    ax21.set_ylabel('Cons Qty')

    ax22.plot(energy_cp2k1[:, 1], energy_cp2k1[:, 3],
              lw=4, ls='-', marker='o')
    ax22.plot(energy_cp2k2[:, 1], energy_cp2k2[:, 3],
              lw=2, ls='--', marker='^')
    ax22.set_ylabel('Temperature')

    fig1.tight_layout()
    plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    PLOT = len(sys.argv) > 3
    main(sys.argv[1], sys.argv[2], plot=PLOT)
